"""
Module with an implementation of the 2D cantilever beam reliability problem.

The 2D cantilever beam problem is a reliability test function from [1].
This is an often revisited problem in reliability analysis ([2], [3]).

The problem consists of a cantilever beam with a rectangular cross-section
subjected to uniformly distributed loading.
The maximum deflection at the free end is taken to be the performance
criterion.


References
----------
1. Malur R. Rajashekhar and Bruce R. Ellingwood, “A new look at the response
   surface approach for reliability analysis,” Structural Safety,
   vol. 12, no. 3, pp. 205–220, 1993.
   DOI: 10.1016/0167-4730(93)90003-J
2. Luc Schueremans and Dionys Van Gemert, “Benefit of splines and neural
   networks in simulation based structural reliability analysis,” Structural
   Safety, vol. 27, no. 3, pp. 246–261, 2005.
   DOI: 10.1016/j.strusafe.2004.11.001
3. X. Li, C. Gong, L. Gu, W. Gao, Z. Jing, and H. Su, “A sequential surrogate
   method for reliability analysis based on radial basis function,”
   Structural Safety, vol. 73, pp. 42–53, 2018.
   DOI: 10.1016/j.strusafe.2018.02.005
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs, FunParamSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["CantileverBeam2D"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Rajashekhar1993": {
        "function_id": "Cantilever2D",
        "description": (
            "Input model for the cantilever beam problem "
            "from Rajashekhar and Ellingwood (1993)"
        ),
        "marginals": [
            {
                "name": "W",
                "distribution": "normal",
                "parameters": [1000.0, 200.0],
                "description": "Load per unit area [N/m^2]",
            },
            {
                "name": "H",
                "distribution": "normal",
                "parameters": [250.0, 37.5],
                "description": "Depth of the cross-section [mm]",
            },
        ],
        "copulas": None,
    },
}


AVAILABLE_PARAMETERS: FunParamSpecs = {
    "Rajashekhar1993": {
        "function_id": "CantileverBeam2D",
        "description": (
            "Parameter set for the 2D cantilever beam problem from "
            "Rajashekhar and Ellingwood (1993)"
        ),
        "declared_parameters": [
            {
                "keyword": "modulus",
                "value": 2.6e4,
                "type": float,
                "description": "Modulus of elasticity 'E' [MPa]",
            },
            {
                "keyword": "span",
                "value": 6.0e3,
                "type": float,
                "description": "Span of the beam 'l' [mm]",
            },
        ],
    },
}


def evaluate(xx: np.ndarray, modulus: float, span: float) -> np.ndarray:
    """Evaluate the 2D cantilever beam function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        A two-dimensional input values given by an N-by-2 array
        where N is the number of input values.
    modulus : float
        The modulus of elasticity in [MPa].
    span : float
        The span of the beam in [mm].

    Returns
    -------
    np.ndarray
        The performance function (serviceability) of the cantilever beam
        system. If negative, the system is in failed state.
        The output is a one-dimensional array of length N.
    """
    modulus *= 1e6  # from [MPa] to [Pa]

    # Get the input
    ww = xx[:, 0]  # Load per unit area [N/m^2] ([Pa])
    hh = xx[:, 1]  # Depth of the cross-section [mm]

    # Compute the performance function
    yy = span / 325 - 12 / 8 * span**4 / modulus * ww / hh**3

    return yy


class CantileverBeam2D(UQTestFunFixDimABC):
    """Concrete implementation of the 2D cantilever beam reliability."""

    _tags = ["reliability"]
    _description = (
        "Cantilever beam reliability problem "
        "from Rajashekhar and Ellington (1993)"
    )
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = AVAILABLE_PARAMETERS

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Core module that contains custom types used in UQTestFuns.

Custom types are used to assist type checker during the code development.
"""

from typing import Callable, List, Union, Optional, Any, Dict, Sequence
from typing_extensions import TypedDict

from uqtestfuns.core.prob_input.marginal import Marginal

__all__ = [
    "MarginalSpec",
    "MarginalSpecs",
    "ProbInputSpecs",
    "ProbInputArgs",
    "FunParamSpecs",
    "FunParamsArgs",
    "DeclaredParameters",
]


class MarginalSpec(TypedDict):
    name: str
    distribution: str
    parameters: List[Union[float, int]]
    description: Optional[str]


MarginalSpecs = List[MarginalSpec]


class ProbInputSpec(TypedDict):
    function_id: str
    description: str
    marginals: Union[List[MarginalSpec], Callable]
    copulas: Optional[Any]


ProbInputSpecs = Dict[str, ProbInputSpec]


class ProbInputArgs(TypedDict):
    function_id: str
    input_id: str
    description: str
    marginals: Sequence[Marginal]
    copulas: Optional[Any]


class DeclaredParameter(TypedDict):
    keyword: str
    value: Any
    type: Optional[type]
    description: Optional[str]


DeclaredParameters = List[DeclaredParameter]


class FunParamsSpec(TypedDict):
    function_id: str
    description: str
    declared_parameters: DeclaredParameters


FunParamSpecs = Dict[str, FunParamsSpec]


class FunParamsArgs(TypedDict):
    function_id: str
    parameter_id: str
    description: str
    declared_parameters: DeclaredParameters



"""
Module with routines involving the normal (Gaussian) probability distribution.

The normal distribution in UQTestFuns is parametrized by two parameters:
mu and sigma, the mean and standard deviation, respectively.

The underlying implementation is based on the implementation from scipy.stats.
In the SciPy convention, the mean (mu) corresponds to the ``loc`` parameter,
while the standard deviation (sigma) corresponds to the ``scale`` parameter.
"""

import numpy as np
from scipy.stats import norm

from .utils import verify_param_nums, postprocess_icdf
from ....global_settings import ARRAY_FLOAT

DISTRIBUTION_NAME = "normal"

NUM_PARAMS = 2


def verify_parameters(parameters: ARRAY_FLOAT) -> None:
    """Verify the parameters of a normal distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a normal distribution
        (i.e., the mean and standard deviation).

    Returns
    ------
    None
        The function exits without any return value when nothing is wrong.

    Raises
    ------
    ValueError
        If any of the parameter values are invalid
        or the shapes are inconsistent.
    """
    # Verify overall shape
    verify_param_nums(parameters.size, NUM_PARAMS, DISTRIBUTION_NAME)

    if parameters[1] <= 0.0:
        raise ValueError(
            f"The corresponding standard deviation {parameters[1]}"
            f"must be larger than 0.0!"
        )


def lower(parameters: ARRAY_FLOAT) -> float:
    """Get the lower bound of a normal distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a normal distribution.

    Returns
    -------
    float
        The lower bound of the normal distribution.

    Notes
    -----
    - Strictly speaking, a normal distribution is unbounded on the left.
      However, for numerical reason a lower bound is set.
    - The lower bound of the normal distribution is chosen such that
      the probability mass between the lower and upper bound is at least
      1 - 1e-15.
    """
    # -8.222082216130435 is the quantile values with probability of 1e-16
    # for the standard Normal distribution (mu = 0.0, sigma = 1.0)
    lower_bound = float(-8.222082216130435 * parameters[1] + parameters[0])

    return lower_bound


def upper(parameters: ARRAY_FLOAT) -> float:
    """Get the upper bound of a normal distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a normal distribution

    Returns
    -------
    float
        The upper bound of the normal distribution.

    Notes
    -----
    - Strictly speaking, a normal distribution is unbounded on the right.
      However, for numerical reason an upper bound is set.
    - The upper bound of the normal distribution is chosen such that
      tbe probability mass between the lower and upper bound is at least
      1 - 1e-15.
    """
    # 8.209536151601387 is the quantile values with probability of 1-1e-16
    # for the standard Normal distribution (mu = 0.0, sigma = 1.0)
    upper_bound = float(8.209536151601387 * parameters[1] + parameters[0])

    return upper_bound


def pdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the PDF values of a normal distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) of a normal distribution.
    parameters : ARRAY_FLOAT
        Parameters of the normal distribution.
    lower_bound : float
        Lower bound of the normal distribution.
    upper_bound : float
        Upper bound of the normal distribution.

    Returns
    -------
    np.ndarray
        PDF values of the normal distribution on the sample values.

    Notes
    -----
    - The PDF for sample with values outside the bounds are set to 0.0.
    """
    yy = np.zeros(xx.shape)
    idx = np.logical_and(xx >= lower_bound, xx <= upper_bound)
    yy[idx] = norm.pdf(xx[idx], loc=parameters[0], scale=parameters[1])

    return yy


def cdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the CDF values of a normal distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) of a normal distribution.
    parameters : ARRAY_FLOAT
        Parameters of the normal distribution.
    lower_bound : float
        Lower bound of the normal distribution.
    upper_bound : float
        Upper bound of the normal distribution.

    Returns
    -------
    ARRAY_FLOAT
        CDF values of the normal distribution on the sample values.

    Notes
    -----
    - The CDF for sample with values smaller (resp. larger)
      than the lower bound (resp. upper bound) are set to 0.0 (resp. 1.0).
    """
    yy = np.empty(xx.shape)
    idx_lower = xx < lower_bound
    idx_upper = xx > upper_bound
    idx_rest = np.logical_and(
        np.logical_not(idx_lower), np.logical_not(idx_upper)
    )

    yy[idx_lower] = 0.0
    yy[idx_upper] = 1.0
    yy[idx_rest] = norm.cdf(
        xx[idx_rest], loc=parameters[0], scale=parameters[1]
    )

    return yy


def icdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the inverse CDF values of a normal distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) in the [0, 1] domain.
    parameters : ARRAY_FLOAT
        Parameters of a normal distribution.
    lower_bound : float
        Lower bound of the normal distribution.
    upper_bound : float
        Upper bound of the normal distribution.

    Returns
    -------
    ARRAY_FLOAT
        Transformed values in the domain of the normal distribution.

    Notes
    -----
    - ICDF for sample values outside [0.0, 1.0] is set to NaN according to
      the underlying SciPy implementation.
    """

    # Compute the ICDF
    yy = norm.ppf(xx, loc=parameters[0], scale=parameters[1])

    # Check if values are within the set bounds
    yy = postprocess_icdf(yy, lower_bound, upper_bound)

    return yy



"""
Module with an implementation of the linear fun. from Saltelli et al. (2008).

The linear function from Saltelli et al. (2008) is an M-dimensional
scalar-valued function.
It was introduced in [1] to illustrate sensitivity analysis methods and used
in [2] for benchmarking purposes.

Due to its simple structure, its moments and Sobol' sensitivity indices can
be computed analytically.

References
----------
1. A. Saltelli, K. Chan, and E. Scott, 2008. Sensitivity Analysis: Wiley Series
   in Probability and Statistics. Wiley, New York.
2. X. Sun, B. Croke, A. Jakeman, S. Roberts, "Benchmarking Active Subspace
   methods of global sensitivity analysis against variance-based Sobol’
   and Morris methods with established test functions," Environmental Modelling
   & Software, vol. 149, p. 105310, 2022.
   DOI: 10.1016/j.envsoft.2022.105310
"""

import numpy as np

from uqtestfuns.core.custom_typing import (
    MarginalSpec,
    MarginalSpecs,
    ProbInputSpecs,
)
from uqtestfuns.core.uqtestfun_abc import UQTestFunVarDimABC

__all__ = ["SaltelliLinear"]


def _create_marginals(input_dimension: int) -> MarginalSpecs:
    r"""Create a list of marginal specifications of the given dimension.

    .. math::

      X_i \sim \mathcal{U}[x_{o, i} - \sigma_{o, i}, x_{o, i} + \sigma_{o, i}]

    where :math:`x_{o, i} = 3^{i - 1}` and :math:`\sigma_{o, i} = 0.5 x_{o, i}`
    for :math:`i = 1, 2, \ldots`.

    Parameters
    ----------
    input_dimension : int
        The number of marginal specifications in the list.

    Returns
    -------
    MarginalSpecs
        The list of marginal specifications.
    """
    marginals = []
    for i in range(input_dimension):
        mid = 3**i
        delta = 0.5 * mid
        marginal: MarginalSpec = {
            "name": f"X{i + 1}",
            "distribution": "uniform",
            "parameters": [mid - delta, mid + delta],
            "description": None,
        }
        marginals.append(marginal)

    return marginals


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Saltelli2000": {
        "function_id": "SaltelliLinear",
        "description": (
            "Probabilistic input model for the linear function "
            "from Saltelli et al. (2000)"
        ),
        "marginals": _create_marginals,
        "copulas": None,
    },
}


def evaluate(xx: np.ndarray):
    """Evaluate the linear function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        M-Dimensional input values given by an N-by-M array where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = np.sum(xx, axis=1)

    return yy


class SaltelliLinear(UQTestFunVarDimABC):
    """An implementation of the M-dimensional Saltelli linear function."""

    _tags = ["sensitivity"]
    _description = "Linear function from Saltelli et al. (2000)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with routines involving the logit-normal probability distribution.

A logit-normal random variable is a random variable whose logit transformation
is a normal random variable.

The logit-normal distribution in UQTestFuns is parametrized by two parameters:
mu and sigma, the mean and standard deviation of the underlying normal
distribution, respectively.

The underlying implementation is based on the implementation from scipy.stats.
In the SciPy implementation, the mean (mu) corresponds
to the ``loc`` parameter, while the standard deviation (sigma) corresponds
to the ``scale`` parameter.
"""

import numpy as np
from scipy.stats import norm
from scipy.special import logit
from scipy.special import expit as logistic

from .utils import verify_param_nums, postprocess_icdf
from ....global_settings import ARRAY_FLOAT

DISTRIBUTION_NAME = "logitnormal"

NUM_PARAMS = 2


def verify_parameters(parameters: ARRAY_FLOAT) -> None:
    """Verify the parameters of a logit-normal distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a logit-normal distribution
        (i.e., the mean and standard deviation).

    Returns
    ------
    None
        The function exits without any return value when nothing is wrong.

    Raises
    ------
    ValueError
        If any of the parameter values are invalid
        or the shapes are inconsistent.
    """
    # Verify overall shape
    verify_param_nums(parameters.size, NUM_PARAMS, DISTRIBUTION_NAME)

    if parameters[1] <= 0.0:
        raise ValueError(
            f"The corresponding standard deviation {parameters[1]}"
            f"must be larger than 0.0!"
        )


def lower(parameters: ARRAY_FLOAT) -> float:
    """Get the lower bound of a logit-normal distribution.

    Parameters
    ----------
    parameters : np.ndarray
        The parameters of a logit-normal distribution (not used).

    Returns
    -------
    float
        The lower bound of the logit-normal distribution.

    Notes
    -----
    - A logit-normal distribution is bounded on the left at 0.0.
    - The parameters of the logit-normal distribution is part of the function
      parameterization for interface consistency.
    """
    lower_bound = 0.0

    return lower_bound


def upper(parameters: ARRAY_FLOAT) -> float:
    """Get the upper bound of a logit-normal distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a logit-normal distribution

    Returns
    -------
    float
        The upper bound of the logit-normal distribution.

    Notes
    -----
    - A logit-normal distribution is bounded on the left at 1.0.
    - The parameters of the logit-normal distribution is part of the function
      parameterization for interface consistency.
    """
    upper_bound = 1.0

    return upper_bound


def pdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the PDF values of a logit-normal distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) of a logit-normal distribution.
    parameters : ARRAY_FLOAT
        Parameters of the logit-normal distribution.
    lower_bound : float
        Lower bound of the logit-normal distribution.
    upper_bound : float
        Upper bound of the logit-normal distribution.

    Returns
    -------
    ARRAY_FLOAT
        PDF values of the logit-normal distribution on the sample values.

    Notes
    -----
    - The PDF for sample with values outside the bounds are set to 0.0.
    """
    xx_trans = logit(xx)
    yy = np.zeros(xx.shape)
    idx = np.logical_and(xx > 0.0, xx < 1.0)
    yy[idx] = (
        norm.pdf(xx_trans[idx], loc=parameters[0], scale=parameters[1])
        / xx[idx]
        / (1 - xx[idx])
    )

    return yy


def cdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the CDF values of a logit-normal distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) of a logit-normal distribution.
    parameters : ARRAY_FLOAT
        Parameters of the logit-normal distribution.
    lower_bound : float
        Lower bound of the logit-normal distribution.
    upper_bound : float
        Upper bound of the logit-normal distribution.

    Returns
    -------
    ARRAY_FLOAT
        CDF values of the logit-normal distribution on the sample values.

    Notes
    -----
    - The CDF for sample with values smaller (resp. larger)
      than the lower bound (resp. upper bound) are set to 0.0 (resp. 1.0).
    """
    xx_trans = logit(xx)
    yy = np.empty(xx.shape)
    idx_lower = xx < lower_bound
    idx_upper = xx > upper_bound
    idx_rest = np.logical_and(
        np.logical_not(idx_lower), np.logical_not(idx_upper)
    )

    yy[idx_lower] = 0.0
    yy[idx_upper] = 1.0
    yy[idx_rest] = norm.cdf(
        xx_trans[idx_rest], loc=parameters[0], scale=parameters[1]
    )

    return yy


def icdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the inverse CDF values of a logit-normal distribution.

    Parameters
    ----------
    xx : np.ndarray
        Sample values (realizations) in the [0, 1] domain.
    parameters : np.ndarray
        Parameters of a logit-normal distribution.
    lower_bound : np.ndarray
        Lower bound of the logit-normal distribution.
    upper_bound : np.ndarray
        Upper bound of the logit-normal distribution.

    Returns
    -------
    np.ndarray
        Transformed values in the domain of the logit-normal distribution.

    Notes
    -----
    - ICDF for sample values outside [0.0, 1.0] is set to NaN according to
      the underlying SciPy implementation.
    """

    # Compute the ICDF
    yy = logistic(norm.ppf(xx, loc=parameters[0], scale=parameters[1]))

    # Check if values are within the set bounds
    yy = postprocess_icdf(yy, lower_bound, upper_bound)

    return yy



"""
Module with an implementation of the OTL circuit test function.

The OTL circuit test function is a six-dimensional scalar-valued function
that computes the mid-point voltage
of an output transformerless (OTL) push-pull circuit.
The function has been used as a test function in metamodeling exercises [1].
A 20-dimensional variant was used for sensitivity analysis in [2]
by introducing 14 additional inert input variables.

References
----------

1. E. N. Ben-Ari and D. M. Steinberg, "Modeling data from computer
   experiments: An empirical comparison of Kriging with MARS
   and projection pursuit regression," Quality Engineering,
   vol. 19, pp. 327-338, 2007.
   DOI: 10.1080/08982110701580930
2. H. Moon, "Design and Analysis of Computer Experiments for Screening Input
   Variables," Ph.D. dissertation, Ohio State University, Ohio, 2010.
   URL: http://rave.ohiolink.edu/etdc/view?acc_num=osu1275422248
"""

import numpy as np

from copy import deepcopy

from uqtestfuns.core.custom_typing import MarginalSpecs, ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["OTLCircuit"]


MARGINALS_BENARI2007: MarginalSpecs = [
    {
        "name": "Rb1",
        "distribution": "uniform",
        "parameters": [50.0, 150.0],
        "description": "Resistance b1 [kOhm]",
    },
    {
        "name": "Rb2",
        "distribution": "uniform",
        "parameters": [25.0, 70.0],
        "description": "Resistance b2 [kOhm]",
    },
    {
        "name": "Rf",
        "distribution": "uniform",
        "parameters": [0.5, 3.0],
        "description": "Resistance f [kOhm]",
    },
    {
        "name": "Rc1",
        "distribution": "uniform",
        "parameters": [1.2, 2.5],
        "description": "Resistance c1 [kOhm]",
    },
    {
        "name": "Rc2",
        "distribution": "uniform",
        "parameters": [0.25, 1.20],
        "description": "Resistance c2 [kOhm]",
    },
    {
        "name": "beta",
        "distribution": "uniform",
        "parameters": [50.0, 300.0],
        "description": "Current gain [A]",
    },
]

MARGINALS_MOON2010 = [deepcopy(_) for _ in MARGINALS_BENARI2007]
for i in range(14):
    MARGINALS_MOON2010.append(
        {
            "name": f"Inert {i+1}",
            "distribution": "uniform",
            "parameters": [100.0, 200.0],
            "description": "Inert input [-]",
        },
    )

AVAILABLE_INPUTS: ProbInputSpecs = {
    "BenAri2007": {
        "function_id": "OTLCircuit",
        "description": (
            "Probabilistic input model for the OTL Circuit function "
            "from Ben-Ari and Steinberg (2007)."
        ),
        "marginals": MARGINALS_BENARI2007,
        "copulas": None,
    },
    "Moon2010": {
        "function_id": "OTLCircuit",
        "description": (
            "Probabilistic input model for the OTL Circuit function "
            "from Moon (2010)."
        ),
        "marginals": MARGINALS_MOON2010,
        "copulas": None,
    },
}

DEFAULT_INPUT_SELECTION = "BenAri2007"


def evaluate(xx: np.ndarray) -> np.ndarray:
    """Evaluate the OTL circuit test function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        (At least) 6-dimensional input values given by N-by-6 arrays
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the OTL circuit test function,
        i.e., the mid-point voltage in Volt.
        The output is a one-dimensional array of length N.

    Notes
    -----
    - The variant of this test function has 14 additional inputs,
      but they are all taken to be inert and therefore should not affect
      the output.
    """
    rr_b1 = xx[:, 0]  # Resistance b1
    rr_b2 = xx[:, 1]  # Resistance b2
    rr_f = xx[:, 2]  # Resistance f
    rr_c1 = xx[:, 3]  # Resistance c1
    rr_c2 = xx[:, 4]  # Resistance c2
    beta = xx[:, 5]  # Current gain

    # Compute the voltage across b1
    vb1 = 12 * rr_b2 / (rr_b1 + rr_b2)

    # Compute the mid-point voltage
    denom = beta * (rr_c2 + 9) + rr_f
    term_1 = ((vb1 + 0.74) * beta * (rr_c2 + 9)) / denom
    term_2 = 11.35 * rr_f / denom
    term_3 = 0.74 * rr_f * beta * (rr_c2 + 9) / (rr_c1 * denom)

    vm = term_1 + term_2 + term_3

    return vm


class OTLCircuit(UQTestFunFixDimABC):
    """A concrete implementation of the OTL circuit test function."""

    _tags = ["metamodeling", "sensitivity"]
    _default_input_dimension = 6
    _description = (
        "Output transformerless (OTL) circuit model "
        "from Ben-Ari and Steinberg (2007)"
    )
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None
    _default_input_id = DEFAULT_INPUT_SELECTION

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with an implementation of undamped oscillator test function.

The undamped oscillator function is a six-dimensional, scalar-valued test
function that models a non-linear, undamped, single-degree-of-freedom, forced
oscillating mechanical system.
This function is frequently used as a test function for reliability analysis
methods (see [1] through [6]). Additionally, in [7], the function is employed
as a test function for metamodeling exercises.


1. C. G. Bucher and U. Bourgund, “A fast and efficient response surface
   approach for structural reliability problems,” Structural Safety, vol. 7,
   no. 1, pp. 57–66, 1990.
   DOI: 10.1016/0167-4730(90)90012-E
2. M. R. Rajashekhar and B. R. Ellingwood, “A new look at the response surface
   approach for reliability analysis,” Structural Safety, vol. 12, no. 3,
   pp. 205–220, 1993.
   DOI: 10.1016/0167-4730(93)90003-J
3. N. Gayton, J. M. Bourinet, and M. Lemaire, “CQ2RS: a new statistical
   approach to the response surface method for reliability analysis,”
   Structural Safety, vol. 25, no. 1, pp. 99–121, 2003.
   DOI: 10.1016/S0167-4730(02)00045-0
4. L. Schueremans and D. Van Gemert, “Benefit of splines and neural networks
   in simulation based structural reliability analysis,” Structural Safety,
   vol. 27, no. 3, pp. 246–261, 2005.
   DOI: 10.1016/j.strusafe.2004.11.001
5. B. Echard, N. Gayton, and M. Lemaire, “AK-MCS: An active learning
   reliability method combining Kriging and Monte Carlo Simulation,”
   Structural Safety, vol. 33, no. 2, pp. 145–154, 2011.
   DOI: 10.1016/j.strusafe.2011.01.002.
6. B. Echard, N. Gayton, M. Lemaire, and N. Relun, “A combined Importance
   Sampling and Kriging reliability method for small failure probabilities
   with time-demanding numerical models,” Reliability Engineering &
   System Safety, vol. 111, pp. 232–240, 2013.
   DOI: 10.1016/j.ress.2012.10.008.
8. N. Lüthen, S. Marelli, and B. Sudret, “Sparse Polynomial Chaos Expansions:
   Literature Survey and Benchmark,” SIAM/ASA Journal of Uncertainty
   Quantification, vol. 9, no. 2, pp. 593–649, 2021.
   DOI: 10.1137/20M1315774
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC


__all__ = ["UndampedOscillator"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Gayton2003": {
        "function_id": "UndampedOscillator",
        "description": (
            "Input model for the undamped non-linear oscillator "
            "from Gayton et al. (2003) (Table 9)"
        ),
        "marginals": [
            {
                "name": "m",
                "distribution": "normal",
                "parameters": [1.0, 0.05],
                "description": "Mass",
            },
            {
                "name": "c1",
                "distribution": "normal",
                "parameters": [1.0, 0.10],
                "description": "Spring (1) constant",
            },
            {
                "name": "c2",
                "distribution": "normal",
                "parameters": [0.1, 0.01],
                "description": "Spring (2) constant",
            },
            {
                "name": "r",
                "distribution": "normal",
                "parameters": [0.5, 0.05],
                "description": "Length of restoring force",
            },
            {
                "name": "F1",
                "distribution": "normal",
                "parameters": [1.0, 0.2],
                "description": "Pulse load",
            },
            {
                "name": "t1",
                "distribution": "normal",
                "parameters": [1.0, 0.2],
                "description": "Duration of the pulse load",
            },
        ],
        "copulas": None,
    },
    "Echard2013-1": {
        "function_id": "UndampedOscillator",
        "description": (
            "Input model for the undamped non-linear oscillator "
            "from Echard et al. (2013) (Table 4, F1 = 0.6)"
        ),
        "marginals": [
            {
                "name": "m",
                "distribution": "normal",
                "parameters": [1.0, 0.05],
                "description": "Mass",
            },
            {
                "name": "c1",
                "distribution": "normal",
                "parameters": [1.0, 0.10],
                "description": "Spring (1) constant",
            },
            {
                "name": "c2",
                "distribution": "normal",
                "parameters": [0.1, 0.01],
                "description": "Spring (2) constant",
            },
            {
                "name": "r",
                "distribution": "normal",
                "parameters": [0.5, 0.05],
                "description": "Length of restoring force",
            },
            {
                "name": "F1",
                "distribution": "normal",
                "parameters": [0.6, 0.1],
                "description": "Pulse load",
            },
            {
                "name": "t1",
                "distribution": "normal",
                "parameters": [1.0, 0.2],
                "description": "Duration of the pulse load",
            },
        ],
        "copulas": None,
    },
    "Echard2013-2": {
        "function_id": "UndampedOscillator",
        "description": (
            "Input model for the undamped non-linear oscillator "
            "from Echard et al. (2013) (Table 4, F1 = 0.45)"
        ),
        "marginals": [
            {
                "name": "m",
                "distribution": "normal",
                "parameters": [1.0, 0.05],
                "description": "Mass",
            },
            {
                "name": "c1",
                "distribution": "normal",
                "parameters": [1.0, 0.10],
                "description": "Spring (1) constant",
            },
            {
                "name": "c2",
                "distribution": "normal",
                "parameters": [0.1, 0.01],
                "description": "Spring (2) constant",
            },
            {
                "name": "r",
                "distribution": "normal",
                "parameters": [0.5, 0.05],
                "description": "Length of restoring force",
            },
            {
                "name": "F1",
                "distribution": "normal",
                "parameters": [0.45, 0.45 / 6],
                "description": "Pulse load",
            },
            {
                "name": "t1",
                "distribution": "normal",
                "parameters": [1.0, 0.2],
                "description": "Duration of the pulse load",
            },
        ],
        "copulas": None,
    },
}


def evaluate(xx: np.ndarray) -> np.ndarray:
    """Evaluate the undamped oscillator test function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        A two-dimensional input values given by N-by-6 arrays
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        If negative, then the system is in failed state.
        The output is a 1-dimensional array of length N.
    """
    omega_0 = np.sqrt((xx[:, 1] + xx[:, 2]) / xx[:, 0])
    term_1 = 3 * xx[:, 3]
    term_2 = 2 * xx[:, 4] / xx[:, 0] / omega_0**2
    term_3 = np.sin(omega_0 * xx[:, 5] / 2)

    yy = term_1 - np.abs(term_2 * term_3)

    return yy


class UndampedOscillator(UQTestFunFixDimABC):
    """A concrete implementation of the undamped oscillator test function."""

    _tags = ["reliability", "metamodeling"]
    _description = "Undamped, non-linear, single DOF oscillator"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None
    _default_input_id = "Gayton2003"

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with an implementation of the McLain's test functions.

The McLain's test functions consists of five two-dimensional scalar-valued
functions. The functions were introduced in [1] in the context of drawing
contours from a given set of points.

There are five test functions in McLain's paper each models a mathematically
defined surface:

- S1: A part of a sphere
- S2: A steep hill rising from a plain
- S3: A less steep hill
- S4: A long narrow hill
- S5: A plateau and plain separated by a steep cliff

Four of the functions (S1-S3 and S5) appeared in modified forms in [2].

References
----------

1. D. H. McLain, "Drawing contours from arbitrary data points," The Computer
   Journal, vol. 17, no. 4, pp. 318-324, 1974.
   DOI: 10.1093/comjnl/17.4.318
2. Richard Franke, "A critical comparison of some methods for interpolation
   of scattered data," Naval Postgraduate School, Monterey, Canada,
   Technical Report No. NPS53-79-003, 1979.
   URL: https://core.ac.uk/reader/36727660
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["McLainS1", "McLainS2", "McLainS3", "McLainS4", "McLainS5"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "McLain1974": {
        "function_id": "McLain",
        "description": (
            "Input specification for the McLain's test functions "
            "from McLain (1974)."
        ),
        "marginals": [  # From Ref. [1]
            {
                "name": "X1",
                "distribution": "uniform",
                "parameters": [1.0, 10.0],
                "description": None,
            },
            {
                "name": "X2",
                "distribution": "uniform",
                "parameters": [1.0, 10.0],
                "description": None,
            },
        ],
        "copulas": None,
    },
}

COMMON_METADATA = dict(
    _tags=["metamodeling"],
    _available_inputs=AVAILABLE_INPUTS,
    _available_parameters=None,
    _description="from McLain (1974)",
)


def evaluate_mclain_s1(xx: np.ndarray) -> np.ndarray:
    """Evaluate the McLain S1 function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        Two-Dimensional input values given by N-by-2 arrays where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the McLain S1 function evaluated
        on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = np.sqrt(64 - (xx[:, 0] - 5.5) ** 2 - (xx[:, 1] - 5.5) ** 2)

    return yy


class McLainS1(UQTestFunFixDimABC):
    """A concrete implementation of the McLain S1 function.

    The function features a part of a sphere.
    """

    _tags = COMMON_METADATA["_tags"]
    _description = f"McLain S1 function {COMMON_METADATA['_description']}"
    _available_inputs = COMMON_METADATA["_available_inputs"]
    _available_parameters = COMMON_METADATA["_available_parameters"]

    evaluate = staticmethod(evaluate_mclain_s1)  # type: ignore


def evaluate_mclain_s2(xx: np.ndarray) -> np.ndarray:
    """Evaluate the McLain S2 function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        Two-Dimensional input values given by N-by-2 arrays where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the McLain S2 function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = np.exp(-1.0 * ((xx[:, 0] - 5) ** 2 + (xx[:, 1] - 5) ** 2))

    return yy


class McLainS2(UQTestFunFixDimABC):
    """A concrete implementation of the McLain S2 function.

    The function features a steep hill rising from a plain.
    """

    _tags = COMMON_METADATA["_tags"]
    _description = f"McLain S2 function {COMMON_METADATA['_description']}"
    _available_inputs = COMMON_METADATA["_available_inputs"]
    _available_parameters = COMMON_METADATA["_available_parameters"]

    evaluate = staticmethod(evaluate_mclain_s2)  # type: ignore


def evaluate_mclain_s3(xx: np.ndarray) -> np.ndarray:
    """Evaluate the McLain S3 function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        Two-Dimensional input values given by N-by-2 arrays where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the McLain S3 function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = np.exp(-0.25 * ((xx[:, 0] - 5) ** 2 + (xx[:, 1] - 5) ** 2))

    return yy


class McLainS3(UQTestFunFixDimABC):
    """A concrete implementation of the McLain S3 function.

    The function features a less steep hill (compared to S2).
    """

    _tags = COMMON_METADATA["_tags"]
    _description = f"McLain S3 function {COMMON_METADATA['_description']}"
    _available_inputs = COMMON_METADATA["_available_inputs"]
    _available_parameters = COMMON_METADATA["_available_parameters"]

    evaluate = staticmethod(evaluate_mclain_s3)  # type: ignore


def evaluate_mclain_s4(xx: np.ndarray) -> np.ndarray:
    """Evaluate the McLain S4 function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        Two-Dimensional input values given by N-by-2 arrays where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the McLain S4 function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = np.exp(
        -1
        * ((xx[:, 0] + xx[:, 1] - 11) ** 2 + (xx[:, 0] - xx[:, 1]) ** 2 / 10.0)
    )

    return yy


class McLainS4(UQTestFunFixDimABC):
    """A concrete implementation of the McLain S4 function.

    The function features a long narrow hill.
    """

    _tags = COMMON_METADATA["_tags"]
    _description = f"McLain S4 function {COMMON_METADATA['_description']}"
    _available_inputs = COMMON_METADATA["_available_inputs"]
    _available_parameters = COMMON_METADATA["_available_parameters"]

    evaluate = staticmethod(evaluate_mclain_s4)  # type: ignore


def evaluate_mclain_s5(xx: np.ndarray) -> np.ndarray:
    """Evaluate the McLain S5 function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        Two-Dimensional input values given by N-by-2 arrays where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the McLain S5 function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = np.tanh(xx[:, 0] + xx[:, 1] - 11)

    return yy


class McLainS5(UQTestFunFixDimABC):
    """A concrete implementation of the McLain S5 function.

    The function features two plateaus separated by a steep cliff.
    """

    _tags = COMMON_METADATA["_tags"]
    _description = f"McLain S5 function {COMMON_METADATA['_description']}"
    _available_inputs = COMMON_METADATA["_available_inputs"]
    _available_parameters = COMMON_METADATA["_available_parameters"]

    evaluate = staticmethod(evaluate_mclain_s5)  # type: ignore



"""
This module contains the concrete implementation of a generic class to
create a UQ test function in runtime or within a running Python session.
"""

from typing import Callable, Optional

from .uqtestfun_abc import UQTestFunBareABC
from .parameters import FunParams
from .prob_input.probabilistic_input import ProbInput

__all__ = ["UQTestFun"]


class UQTestFun(UQTestFunBareABC):
    """Generic concrete class of UQ test function.

    Parameters
    ----------
    evaluate : Callable
        The evaluation function of the UQ test function implemented as a
        Python callable.
    prob_input : ProbInput
        The probabilistic input model of the UQ test function.
    parameters : Any, optional
        The parameters set of the UQ test function.
        If not specified, `None` is used.
    function_id : str, optional
        The ID of the UQ test function.
    """

    def __init__(
        self,
        evaluate: Callable,
        prob_input: ProbInput,
        parameters: Optional[FunParams] = None,
        function_id: Optional[str] = None,
    ):
        if parameters is None:
            parameters = FunParams()

        self._evaluate = evaluate
        super().__init__(prob_input, parameters, function_id)

    def _eval(self, xx):
        return self._evaluate(xx, **self.parameters.as_dict())

    evaluate = None  # type: ignore



"""
Module with an implementation of the Wing Weight test function.

The Wing Weight test function [1] is a 10-dimensional scalar-valued function
that models a light aircraft wing.
The function has been used as a test function in the context of
metamodeling [2] and optimization [1].

References
----------

1. Alexander I. J. Forrester, András Sóbester, and Andy J. Keane,
   Engineering Design via Surrogate Modelling: A Practical Guide, 1st ed.
   Wiley, 2008.
   DOI: 10.1002/9780470770801.
2. Lavi R. Zuhal, Kemas Zakaria, Pramudita S. Palar, Koji Shimoyama,
   and Rhea P. Liem, "Gradient-enhanced universal Kriging with polynomial
   chaos as trend function," In AIAA Scitech 2020 Forum,
   Orlando, Florida, 2020. American Institute of Aeronautics and Astronautics.
   DOI: 10.2514/6.2020-1865.
"""

import numpy as np

from uqtestfuns.core.custom_typing import MarginalSpecs, ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC
from .utils import deg2rad

__all__ = ["WingWeight"]


MARGINALS_FORRESTER2008: MarginalSpecs = [
    {
        "name": "Sw",
        "distribution": "uniform",
        "parameters": [150.0, 200.0],
        "description": "wing area [ft^2]",
    },
    {
        "name": "Wfw",
        "distribution": "uniform",
        "parameters": [220.0, 300.0],
        "description": "weight of fuel in the wing [lb]",
    },
    {
        "name": "A",
        "distribution": "uniform",
        "parameters": [6.0, 10.0],
        "description": "aspect ratio [-]",
    },
    {
        "name": "Lambda",
        "distribution": "uniform",
        "parameters": [-10.0, 10.0],
        "description": "quarter-chord sweep [degrees]",
    },
    {
        "name": "q",
        "distribution": "uniform",
        "parameters": [16.0, 45.0],
        "description": "dynamic pressure at cruise [lb/ft^2]",
    },
    {
        "name": "lambda",
        "distribution": "uniform",
        "parameters": [0.5, 1.0],
        "description": "taper ratio [-]",
    },
    {
        "name": "tc",
        "distribution": "uniform",
        "parameters": [0.08, 0.18],
        "description": "aerofoil thickness to chord ratio [-]",
    },
    {
        "name": "Nz",
        "distribution": "uniform",
        "parameters": [2.5, 6.0],
        "description": "ultimate load factor [-]",
    },
    {
        "name": "Wdg",
        "distribution": "uniform",
        "parameters": [1700, 2500],
        "description": "flight design gross weight [lb]",
    },
    {
        "name": "Wp",
        "distribution": "uniform",
        "parameters": [0.025, 0.08],
        "description": "paint weight [lb/ft^2]",
    },
]

AVAILABLE_INPUTS: ProbInputSpecs = {
    "Forrester2008": {
        "function_id": "WingWeight",
        "description": (
            "Probabilistic input model for the Wing Weight model "
            "from Forrester et al. (2008)."
        ),
        "marginals": MARGINALS_FORRESTER2008,
        "copulas": None,
    },
}


def evaluate(xx: np.ndarray) -> np.ndarray:
    """Evaluate the Wing Weight function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        10-Dimensional input values given by N-by-10 arrays where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the Wing Weight function evaluated
        on the input values.
        The output is a 1-dimensional array of length N.
    """
    # Compute the Wing Weight function
    term_1 = 0.036 * xx[:, 0] ** 0.758 * xx[:, 1] ** 0.0035
    term_2 = (xx[:, 2] / np.cos(deg2rad(xx[:, 3])) ** 2) ** 0.6
    term_3 = xx[:, 4] ** 0.006
    term_4 = xx[:, 5] ** 0.04
    term_5 = (100 * xx[:, 6] / np.cos(np.pi / 180.0 * xx[:, 3])) ** (-0.3)
    term_6 = (xx[:, 7] * xx[:, 8]) ** 0.49
    term_7 = xx[:, 0] * xx[:, 9]

    yy = term_1 * term_2 * term_3 * term_4 * term_5 * term_6 + term_7

    return yy


class WingWeight(UQTestFunFixDimABC):
    """A concrete implementation of the wing weight test function."""

    _tags = ["metamodeling", "sensitivity"]
    _description = "Wing weight model from Forrester et al. (2008)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with an implementation of the M-dimensional Ackley function.

The Ackley function  is an $M$-dimensional scalar-valued function.
The function was first introduced in [1] as a test function for optimization
algorithms,
Originally presented as a two-dimensional function, it was later generalized
by Bäck and Schwefel [2].

References
----------

1. D. H. Ackley, A Connectionist Machine for Genetic Hillclimbing,
   The Kluwer International Series in Engineering and Computer Science vol. 28.
   Boston, MA: Springer US, 1987.
   DOI: 10.1007/978-1-4613-1997-9
2. T. Bäck and H.-P. Schwefel, “An overview of evolutionary algorithms for
   parameter optimization,” Evolutionary Computation,
   vol. 1, no. 1, pp. 1–23, 1993.
   DOI: 10.1162/evco.1993.1.1.1.
"""

import numpy as np

from uqtestfuns.core.uqtestfun_abc import UQTestFunVarDimABC
from uqtestfuns.core.custom_typing import ProbInputSpecs, FunParamSpecs

__all__ = ["Ackley"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Ackley1987": {
        "function_id": "Ackley",
        "description": (
            "Search domain for the Ackley function from Ackley (1987)"
        ),
        "marginals": [
            {
                "name": "X",
                "distribution": "uniform",
                "parameters": [-32.768, 32.768],
                "description": None,
            },
        ],
        "copulas": None,
    }
}


AVAILABLE_PARAMETERS: FunParamSpecs = {
    "Ackley1987": {
        "function_id": "Ackley",
        "description": (
            "Parameter set for the Ackley function from Ackley (1987)"
        ),
        "declared_parameters": [
            {
                "keyword": "a",
                "value": 20.0,
                "type": float,
                "description": "Height of the ridges surrounding the minimum",
            },
            {
                "keyword": "b",
                "value": 0.2,
                "type": float,
                "description": "Decay rate of the Euclidean distance",
            },
            {
                "keyword": "c",
                "value": 2 * np.pi,
                "type": float,
                "description": "Scaling constant for the cosine term",
            },
        ],
    },
}


def evaluate(xx: np.ndarray, a: float, b: float, c: float) -> np.ndarray:
    """Evaluate the Ackley function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        M-Dimensional input values given by an N-by-M array where
        N is the number of input values.
    a : float
        Height of the ridges surrounding the minimum.
    b : float
        Decay rate of the Euclidean distance.
    c : float
        Scaling constant for the cosine term.

    Returns
    -------
    np.ndarray
        The output of the Ackley function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """

    m = xx.shape[1]

    # Compute the Ackley function
    term_1 = -1 * a * np.exp(-1 * b * np.sqrt(np.sum(xx**2, axis=1) / m))
    term_2 = -1 * np.exp(np.sum(np.cos(c * xx), axis=1) / m)

    yy = term_1 + term_2 + a + np.exp(1)

    return yy


class Ackley(UQTestFunVarDimABC):
    """A concrete implementation of the M-dimensional Ackley test function."""

    _tags = ["optimization", "metamodeling"]
    _description = "Optimization test function from Ackley (1987)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = AVAILABLE_PARAMETERS

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with an implementation of the Borehole function.

The Borehole test function [1] is an eight-dimensional scalar-valued function
that models water flow through a borehole that is drilled from
the ground surface through two aquifers.

The unit of the output is [m^3/year].

References
----------

1. W. V. Harper and S. K. Gupta, “Sensitivity/Uncertainty Analysis of
   a Borehole Scenario Comparing Latin Hypercube Sampling and
   Deterministic Sensitivity Approaches”, Office of Nuclear Waste
   Isolation, Battelle Memorial Institute, Columbus, Ohio,
   BMI/ONWI-516, 1983.
   URL: https://inldigitallibrary.inl.gov/PRR/84393.pdf
2. Max D. Morris, T. J. Mitchell, and D. Ylvisaker, “Bayesian design and
   analysis of computer experiments: Use of derivatives in surface
   prediction,” Technometrics, vol. 35, no. 3, pp. 243–255, 1993.
   DOI: 10.1080/00401706.1993.10485320
"""

import numpy as np

from uqtestfuns.core.custom_typing import MarginalSpecs, ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["Borehole"]


# From Ref. [1]
MARGINALS_HARPER1983: MarginalSpecs = [
    {
        "name": "rw",
        "distribution": "normal",
        "parameters": [0.10, 0.0161812],
        "description": "radius of the borehole [m]",
    },
    {
        "name": "r",
        "distribution": "lognormal",
        "parameters": [7.71, 1.0056],
        "description": "radius of influence [m]",
    },
    {
        "name": "Tu",
        "distribution": "uniform",
        "parameters": [63070.0, 115600.0],
        "description": "transmissivity of upper aquifer [m^2/year]",
    },
    {
        "name": "Hu",
        "distribution": "uniform",
        "parameters": [990.0, 1100.0],
        "description": "potentiometric head of upper aquifer [m]",
    },
    {
        "name": "Tl",
        "distribution": "uniform",
        "parameters": [63.1, 116.0],
        "description": "transmissivity of lower aquifer [m^2/year]",
    },
    {
        "name": "Hl",
        "distribution": "uniform",
        "parameters": [700.0, 820.0],
        "description": "potentiometric head of lower aquifer [m]",
    },
    {
        "name": "L",
        "distribution": "uniform",
        "parameters": [1120.0, 1680.0],
        "description": "length of the borehole [m]",
    },
    {
        "name": "Kw",
        "distribution": "uniform",
        "parameters": [9985.0, 12045.0],
        "description": "hydraulic conductivity of the borehole [m/year]",
    },
]

# From Ref. [2]
MARGINALS_MORRIS1993 = list(MARGINALS_HARPER1983)
MARGINALS_MORRIS1993[0:2] = [
    {
        "name": "rw",
        "distribution": "uniform",
        "parameters": [0.05, 0.15],
        "description": "radius of the borehole [m]",
    },
    {
        "name": "r",
        "distribution": "uniform",
        "parameters": [100, 50000],
        "description": "radius of influence [m]",
    },
]

AVAILABLE_INPUTS: ProbInputSpecs = {
    "Harper1983": {
        "function_id": "Borehole",
        "description": (
            "Probabilistic input model of the Borehole model "
            "from Harper and Gupta (1983)"
        ),
        "marginals": MARGINALS_HARPER1983,
        "copulas": None,
    },
    "Morris1993": {
        "function_id": "Borehole",
        "description": (
            "Probabilistic input model of the Borehole model "
            "from Morris et al. (1993)"
        ),
        "marginals": MARGINALS_MORRIS1993,
        "copulas": None,
    },
}

DEFAULT_INPUT_SELECTION = "Harper1983"


def evaluate(xx: np.ndarray) -> np.ndarray:
    """Evaluate the Borehole function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        8-Dimensional input values given by N-by-8 arrays where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the Borehole function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    # Compute the Borehole function
    nom = 2 * np.pi * xx[:, 2] * (xx[:, 3] - xx[:, 5])
    denom_1 = np.log(xx[:, 1] / xx[:, 0])
    denom_2 = (
        2
        * xx[:, 6]
        * xx[:, 2]
        / (np.log(xx[:, 1] / xx[:, 0]) * xx[:, 0] ** 2 * xx[:, 7])
    )
    denom_3 = xx[:, 2] / xx[:, 4]

    yy = nom / (denom_1 * (1 + denom_2 + denom_3))

    return yy


class Borehole(UQTestFunFixDimABC):
    """A concrete implementation of the Borehole function."""

    _tags = ["metamodeling", "sensitivity"]
    _description = "Borehole function from Harper and Gupta (1983)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None
    _default_input_id = DEFAULT_INPUT_SELECTION

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with an implementation of the Friedman test functions.

The Friedman function is a function introduced in [2] as a six-dimensional
(including one dummy variable) test function for testing a spline approximation
method. It was later modified to be a ten-dimensional (including five dummy
variables) function in [2].
In [3], the function was employed as a test function in the context of
sensitivity analysis.

References
----------

1. J. H. Friedman, E. Grosse, and W. Stuetzle, "Multidimensional additive
   spline approximation," SIAM Journal on Scientific and Statistical Computing,
   vol. 4, no. 2, pp. 291-301, 1983.
2. J. H. Friedman, "Multivariate adaptive regression splines," The Annals of
   Statistics, vol. 19, no. 1, pp. 1-67, 1991.
   DOI: 10.1214/aos/1176347963
3. X. Sun, B. Croke, A. Jakeman, S. Roberts, "Benchmarking Active Subspace
   methods of global sensitivity analysis against variance-based Sobol’
   and Morris methods with established test functions," Environmental
   Modelling & Software, vol. 149, p. 105310, 2022.
   DOI: 10.1016/j.envsoft.2022.105310
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["Friedman6D", "Friedman10D"]


AVAILABLE_INPUTS_6D: ProbInputSpecs = {
    "Friedman1983": {
        "function_id": "Friedman6D",
        "description": (
            "Input specification for the 6D test function "
            "from Friedman et al. (1983)"
        ),
        "marginals": [
            {
                "name": f"x_{i + 1}",
                "distribution": "uniform",
                "parameters": [0, 1],
                "description": None,
            }
            for i in range(6)
        ],
        "copulas": None,
    },
}


AVAILABLE_INPUTS_10D: ProbInputSpecs = {
    "Friedman1983": {
        "function_id": "Friedman10D",
        "description": (
            "Input specification for the 6D test function "
            "from Friedman (1991)"
        ),
        "marginals": [
            {
                "name": f"x_{i + 1}",
                "distribution": "uniform",
                "parameters": [0, 1],
                "description": None,
            }
            for i in range(10)
        ],
        "copulas": None,
    },
}


def evaluate_friedman(xx: np.ndarray) -> np.ndarray:
    """Evaluate the six-dimensional Friedman test function.

    Parameters
    ----------
    xx : np.ndarray
        6-Dimensional input values given by an N-by-6 array
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the six-dimensional Friedman function evaluated
        on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy_1 = 10 * np.sin(np.pi * xx[:, 0] * xx[:, 1])
    yy_2 = 20 * (xx[:, 2] - 0.5) ** 2
    yy_3 = 10 * xx[:, 3]
    yy_4 = 5 * xx[:, 5]

    return yy_1 + yy_2 + yy_3 + yy_4


class Friedman6D(UQTestFunFixDimABC):
    """A concrete implementation of the 6D Friedman et al. (1983) function."""

    _tags = ["metamodeling", "sensitivity"]
    _description = "Six-dimensional function from Friedman et al. (1983)"
    _available_inputs = AVAILABLE_INPUTS_6D
    _available_parameters = None

    evaluate = staticmethod(evaluate_friedman)  # type: ignore


class Friedman10D(UQTestFunFixDimABC):
    """A concrete implementation of the 10D Friedman (1991) function."""

    _tags = ["metamodeling"]
    _description = "Ten-dimensional function from Friedman (1991)"
    _available_inputs = AVAILABLE_INPUTS_10D
    _available_parameters = None

    evaluate = staticmethod(evaluate_friedman)  # type: ignore



"""
Module with an implementation of the Sulfur model test function.

The Sulfur model from [1] is a nine-dimensional scalar-valued function
that analytically computes the direct radiative forcing by sulfate aerosols.
The test function was used in [2] for metamodeling and uncertainty propagation
purposes.
The probabilistic input specification used in [2] was originally based
from [3].

Due to the construction of the test function, the geometric mean and geometric
standard deviation of the response are analytically available given the input
specification (from [2], all are lognormals). Note that the geometric mean
and the geometric standard deviation of a lognormal distribution are
the exponentiated mu and sigma parameters (i.e., the mean and
standard deviation of the underlying normal distribution).

To complete the test function specification, two additional parameters are
taken from the literature: the solar constant [4] and the surface area
of the earth [5].

Notes
-----

- The equation of the Sulfur model in [2] (Eq. (12)) is erroneous.
  The solar constant term S0 should not be squared otherwise the dimension
  will not agree. Moreover, the equation is missing factors of 365 in the
  denominator and 10^12 in the numerator because the parameter L
  (Sulfate lifetime in the atmosphere) is given in [days]
  while the parameter Q (Global input flux of anthropogenic sulfur is
  given in [TgS/year].
- The input specification of the model here is taken from [3] (Table 2).
  While they are similar, the parameters T and (1-Rs) in [3] are in the
  squared form (i.e., T^2 and (1-Rs)^2).

References
----------
1. R. J. Charlson, S. E. Schwartz, J. M. Hales, R. D. Cess, J. A. Coakley,
   J. E. Hansen, and D. J. Hofmann, “Climate forcing by anthropogenic
   aerosols,” Science, vol. 255, no. 5043, pp. 423–430, 1992.
   DOI: 10.1126/science.255.5043.423.
2. M. A. Tatang, W. Pan, R. G. Prinn, and G. J. McRae, “An efficient method
   for parametric uncertainty analysis of numerical geophysical models,”
   Journal of Geophysical Research, vol. 102, no. D18, pp. 21925–21932, 1997.
   DOI: 10.1029/97JD01654.
3. J. E. Penner, R. J. Charlson, S.E. Schwartz, J. M. Hales, N. S. Laulainen,
   L. Travis, R. Leifer, T. Novakov, J. Ogren, L. F. Radke, “Quantifying and
   Minimizing Uncertainty of Climate Forcing by Anthropogenic Aerosols,”
   Bulletin of the American Meteorological Society, vol. 75, no. 3,
   pp. 375–400, 1994.
   DOI: 10.1175/1520-0477(1994)075<0375:QAMUOC>2.0.CO;2.
4. G. Kopp and J. L. Lean, “A new, lower value of total solar irradiance:
   Evidence and climate significance,” Geophysical Research Letter, vol. 38,
   no. 1, 2011.
   DOI: 10.1029/2010GL045777.
5. M. Pidwirny, “Introduction to the Oceans,”
   Fundamentals of Physical Geography, 2nd Edition, 2006.
   Accessed: Jan. 25, 2023
   URL: http://www.physicalgeography.net/fundamentals/8o.html.
"""

import numpy as np

from uqtestfuns.core.custom_typing import MarginalSpecs, ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["Sulfur"]


MARGINALS_PENNER1994: MarginalSpecs = [  # From [3] (Table 2)
    {
        "name": "Q",
        "distribution": "lognormal",
        "parameters": [np.log(71.0), np.log(1.15)],
        "description": (
            "Source strength of anthropogenic Sulfur [10^12 g/year]"
        ),
    },
    {
        "name": "Y",
        "distribution": "lognormal",
        "parameters": [np.log(0.5), np.log(1.5)],
        "description": "Fraction of SO2 oxidized to SO4(2-) aerosol [-]",
    },
    {
        "name": "L",
        "distribution": "lognormal",
        "parameters": [np.log(5.5), np.log(1.5)],
        "description": "Average lifetime of atmospheric SO4(2-) [days]",
    },
    {
        "name": "Psi_e",
        "distribution": "lognormal",
        "parameters": [np.log(5.0), np.log(1.4)],
        "description": "Aerosol mass scattering efficiency [m^2/g]",
    },
    {
        "name": "beta",
        "distribution": "lognormal",
        "parameters": [np.log(0.3), np.log(1.3)],
        "description": "Fraction of light scattered upward hemisphere [-]",
    },
    {
        "name": "f_Psi_e",
        "distribution": "lognormal",
        "parameters": [np.log(1.7), np.log(1.2)],
        "description": "Fractional increase in aerosol scattering efficiency "
        "due to hygroscopic growth [-]",
    },
    {
        "name": "T^2",
        "distribution": "lognormal",
        "parameters": [np.log(0.58), np.log(1.4)],
        "description": "Square of atmospheric "
        "transmittance above aerosol layer [-]",
    },
    {
        "name": "(1-Ac)",
        "distribution": "lognormal",
        "parameters": [np.log(0.39), np.log(1.1)],
        "description": "Fraction of earth not covered by cloud [-]",
    },
    {
        "name": "(1-Rs)^2",
        "distribution": "lognormal",
        "parameters": [np.log(0.72), np.log(1.2)],
        "description": "Square of surface coalbedo [-]",
    },
]

AVAILABLE_INPUTS: ProbInputSpecs = {
    "Penner1994": {
        "function_id": "Sulfur",
        "description": (
            "Probabilistic input model for the Sulfur model "
            "from Penner et al. (1994)."
        ),
        "marginals": MARGINALS_PENNER1994,
        "copulas": None,
    },
}

DEFAULT_INPUT_SELECTION = "Penner1994"

SOLAR_CONSTANT = 1361  # [W/m^2] from [4]
EARTH_AREA = 5.1e14  # [m^2] from [5]
DAYS_IN_YEAR = 365  # [days]


def evaluate(xx: np.ndarray) -> np.ndarray:
    """Evaluate the Sulfur model test function on a set of input values.

    References
    ----------
    xx : np.ndarray
        A nine-dimensional input values given by an N-by-9 array
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the Sulfur model test function, i.e.,
        the direct radiative forcing by sulfate aerosols.
    """
    # Source strength of anthropogenic Sulfur (initially given in Teragram)
    qq = xx[:, 0] * 1e12
    # Fraction of SO2 oxidized to SO4(2-) aerosol
    yy = xx[:, 1]
    # Average lifetime of atmospheric SO4(2-)
    ll = xx[:, 2]
    # Aerosol mass scattering efficiency
    psi_e = xx[:, 3]
    # Fraction of light scattered upward hemisphere
    beta = xx[:, 4]
    # Fractional increase in aerosol scattering eff. due hygroscopic growth
    ff_psi = xx[:, 5]
    # Square of atmospheric transmittance above aerosol layer
    tt_sq = xx[:, 6]
    # Fraction of earth not covered by cloud
    aa_c_complement = xx[:, 7]
    # Square of surface coalbedo
    co_rr_s_sq = xx[:, 8]

    # Sulfate burden (Eq. (5) in [1], notation from [2])
    # NOTE: Factor 3.0 due to conversion of mass from S to SO4(2-)
    # NOTE: Factor 1/365.0 due to average lifetime is given in [days]
    #       while qq is in [gS / year]
    sulfate_burden = 3.0 * qq * yy * ll / EARTH_AREA / DAYS_IN_YEAR

    # Loading of sulfate aerosol (Eq. (4) in [1], notation from [2])
    sulfate_loading = psi_e * ff_psi * sulfate_burden

    # Direct radiative forcing by sulfate aerosols
    factor_1 = SOLAR_CONSTANT * aa_c_complement * tt_sq * co_rr_s_sq * beta
    dd_f = -0.5 * factor_1 * sulfate_loading

    return dd_f


class Sulfur(UQTestFunFixDimABC):
    """A concrete implementation of the Sulfur model test function."""

    _tags = ["metamodeling", "sensitivity"]
    _description = "Sulfur model from Charlson et al. (1992)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with routines involving the lognormal probability distribution.

The lognormal distribution in UQTestFuns is parameterized by two parameters:
``mu`` and ``sigma``, the mean and standard deviation of the underlying normal
distribution, respectively.

The underlying implementation is based on the implementation of scipy.stats.
In the SciPy implementation, the lognormal distribution is parameterized by
two parameters: ``s`` and ``scale``, the shape and scaling parameters,
respectively. The shape parameter ``s`` corresponds to the standard deviation
of the underlying normal distribution.

The translation between parameterization of the distribution in UQTestFuns
and SciPy is as follows:

- ``s`` = ``sigma``
- ``scale`` = ``np.exp(mu)``
"""

import numpy as np
from scipy.stats import lognorm

from .utils import verify_param_nums, postprocess_icdf
from ....global_settings import ARRAY_FLOAT

DISTRIBUTION_NAME = "lognormal"

NUM_PARAMS = 2


def verify_parameters(parameters: ARRAY_FLOAT) -> None:
    """Verify the parameters of a lognormal distribution.

    Parameters
    ----------
    parameters : np.ndarray
        The parameters of the lognormal distribution
        (i.e., the mean and standard deviation of the associated normal
        distribution).

    Returns
    ------
    None
        The function exits without any return value when nothing is wrong.

    Raises
    ------
    ValueError
        If any of the parameter values are invalid
        or the shapes are inconsistent.
    """
    # Verify overall shape
    verify_param_nums(parameters.size, NUM_PARAMS, DISTRIBUTION_NAME)

    if parameters[1] <= 0.0:
        raise ValueError(
            f"The corresponding normal standard deviation {parameters[1]}"
            f"must be larger than 0.0!"
        )


def lower(parameters: ARRAY_FLOAT) -> float:
    """Get the lower bound of a lognormal distribution.

    Parameters
    ----------
    parameters : np.ndarray
        The parameters of a lognormal distribution.

    Returns
    -------
    float
        The lower bound of the lognormal distribution.

    Notes
    -----
    - The parameters are not used in determining the lower bound of
      the distribution; it must, however, appear for interface consistency.
      The lower bound of a lognormal distribution is finite; it is 0.0.
    """
    lower_bound = 0.0

    return lower_bound


def upper(parameters: ARRAY_FLOAT) -> float:
    """Get the upper bound of a lognormal distribution.

    Parameters
    ----------
    parameters : np.ndarray
        The parameters of a lognormal distribution.

    Returns
    -------
    float
        The upper bound of the lognormal distribution.

    Notes
    -----
    - Strictly speaking, a lognormal distribution is unbounded on the right.
      However, for numerical reason an upper bound is set.
    - The upper bound of the lognormal distribution is chosen such that
      the probability mass between 0.0 (the lower bound) and the upper bound
      is at least 1 - 1e-15.
    """
    # 8.209536151601387 is the quantile values with probability of 1-1e-16
    # for the corresponding standard Normal dist. (mu = 0.0, beta = 1.0)
    upper_bound = float(
        np.exp(8.209536151601387 * parameters[1] + parameters[0])
    )

    return upper_bound


def pdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the PDF values of a lognormal distribution.

    Parameters
    ----------
    xx : np.ndarray
        Sample values (realizations) of a lognormal distribution.
    parameters : np.ndarray
        Parameters of the lognormal distribution.
    lower_bound : float
        Lower bound of the lognormal distribution.
    upper_bound : float
        Upper bound of the lognormal distribution.

    Returns
    -------
    np.ndarray
        PDF values of the lognormal distribution on the sample values.

    Notes
    -----
    - The values outside the bounds are set to 0.0.
    """
    yy = np.zeros(xx.shape)
    idx = np.logical_and(xx >= lower_bound, xx <= upper_bound)
    yy[idx] = lognorm.pdf(
        xx[idx], s=parameters[1], scale=np.exp(parameters[0])
    )

    return yy


def cdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the CDF values of a lognormal distribution.

    Parameters
    ----------
    xx : np.ndarray
        Sample values (realizations) of a lognormal distribution.
    parameters : np.ndarray
        Parameters of the lognormal distribution.
    lower_bound : float
        Lower bound of the lognormal distribution.
    upper_bound : float
        Upper bound of the lognormal distribution.

    Returns
    -------
    np.ndarray
        CDF values of the lognormal distribution on the sample values.

    Notes
    -----
    - CDF for sample with values smaller (resp. larger) than the lower bound
      (resp. upper bound) are set to 0.0 (resp. 1.0).
    """
    yy = np.empty(xx.shape)
    idx_lower = xx < lower_bound
    idx_upper = xx > upper_bound
    idx_rest = np.logical_and(
        np.logical_not(idx_lower), np.logical_not(idx_upper)
    )

    yy[idx_lower] = 0.0
    yy[idx_upper] = 1.0
    yy[idx_rest] = lognorm.cdf(
        xx[idx_rest], s=parameters[1], scale=np.exp(parameters[0])
    )

    return yy


def icdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the inverse CDF values of a lognormal distribution.

    Parameters
    ----------
    xx : np.ndarray
        Sample values (realizations) in the [0, 1] domain.
    parameters : np.ndarray
        Parameters of a lognormal distribution.
    lower_bound : float
        Lower bound of the lognormal distribution.
    upper_bound : float
        Upper bound of the lognormal distribution.

    Returns
    -------
    np.ndarray
        Transformed values in the domain of the lognormal distribution.

    Notes
    -----
    - ICDF for sample values outside [0.0, 1.0] is set to NaN according to
      the underlying SciPy implementation.
    """

    # Compute the ICDF
    yy = lognorm.ppf(xx, s=parameters[1], scale=np.exp(parameters[0]))

    # Check if values are within the set bounds
    yy = postprocess_icdf(yy, lower_bound, upper_bound)

    return yy



"""
This is the package init for UQTestFuns.
"""

import sys

from .core import Marginal
from .core import ProbInput
from .core import (
    UQTestFunBareABC,
    UQTestFunABC,
    UQTestFunFixDimABC,
    UQTestFunVarDimABC,
)
from .core import UQTestFun
from .core import FunParams


from . import test_functions
from .test_functions import *  # noqa

from .meta import UQMetaFunSpec
from .meta import UQMetaTestFun

from .helpers import list_functions

if sys.version_info >= (3, 8):
    from importlib import metadata
else:  # pragma: no cover
    import importlib_metadata as metadata

__version__ = metadata.version("uqtestfuns")

__all__ = [
    "Marginal",
    "ProbInput",
    "FunParams",
    "UQTestFunBareABC",
    "UQTestFunABC",
    "UQTestFunFixDimABC",
    "UQTestFunVarDimABC",
    "UQTestFun",
    "test_functions",
    "UQMetaFunSpec",
    "UQMetaTestFun",
    "list_functions",
]



"""
Module with routines involving the truncated Gumbel (max.) probability dist.

The truncated Gumbel (max.) distribution in UQTestFuns is parameterized by
four parameters: ``mu``, ``beta``, ``lb``, and ``ub`` that correspond to
the mode, the scale, lower bound, and upper bound, respectively.

The underlying implementation is based on the re-implementation of regular
Gumbel (max.) distribution in UQTestFuns (which in turn, based on scipy.stats).
"""

import numpy as np

from scipy.stats import gumbel_r
from typing import Tuple

from . import gumbel
from .utils import verify_param_nums, postprocess_icdf
from ....global_settings import ARRAY_FLOAT

DISTRIBUTION_NAME = "trunc-gumbel"

NUM_PARAMS = 4


def _get_parameters(
    parameters: ARRAY_FLOAT,
) -> Tuple[float, float, float, float]:
    """Get the parameters of a truncated Gumbel (max.) distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a truncated Gumbel (max.) distribution.

    Returns
    -------
    Tuple[float, float, float, float]
        The mode (mu), scale (beta) lower bound, and upper bound
        of the truncated Gumbel (max.) distribution
        (in that order, as a tuple).
    """
    mu = float(parameters[0])
    beta = float(parameters[1])
    lb = float(parameters[2])
    ub = float(parameters[3])

    return mu, beta, lb, ub


def _compute_normalizing_factor(
    mu: float,
    beta: float,
    lb: float,
    ub: float,
) -> float:
    """Compute the normalizing factor of the truncated distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a truncated Gumbel (max.) distribution.

    Returns
    -------
    float
        The normalizing factor for the truncated distribution.
    """

    lb_quantile = gumbel_r.cdf(lb, loc=mu, scale=beta)
    ub_quantile = gumbel_r.cdf(ub, loc=mu, scale=beta)

    normalizing_factor = ub_quantile - lb_quantile

    return normalizing_factor


def verify_parameters(parameters: ARRAY_FLOAT) -> None:
    """Verify the parameters of a truncated Gumbel (max.) distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a truncated Gumbel (max.) distribution
        (i.e., mu, beta, lower bound, and upper bound).

    Returns
    -------
    None
        The function exits without any return value when nothing is wrong.

    Raises
    ------
    ValueError
        If any of the parameter values are invalid
        or the shape is inconsistent.
    """
    # Verify overall shape
    verify_param_nums(parameters.size, NUM_PARAMS, DISTRIBUTION_NAME)

    mu, beta, lb, ub = _get_parameters(parameters)

    # Verify the first two parameters
    gumbel.verify_parameters(parameters[:2])

    # Check the validity of the parameter values w.r.t the bounds
    if lb >= ub:
        raise ValueError(
            f"The lower bound of a truncated Gumbel (max.) distribution {lb} "
            f"cannot be equal or greater than the upper bound {ub}!"
        )


def lower(parameters: ARRAY_FLOAT) -> float:
    """Get the lower bound of a truncated Gumbel (max.) distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a truncated Gumbel (max.) distribution.

    Returns
    -------
    float
        The lower bound of the truncated Gumbel (max.) distribution.
    """
    _, _, lower_bound, _ = _get_parameters(parameters)

    # Get the upper bound of the untruncated Gumbel distribution
    lb_gumbel = gumbel.lower(parameters[:2])
    if lower_bound < lb_gumbel:
        # Use the corresponding normal lower bound instead
        return lb_gumbel

    return lower_bound


def upper(parameters: ARRAY_FLOAT) -> float:
    """Get the upper bound of a truncated Gumbel (max.) distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of a truncated Gumbel (max.) distribution.

    Returns
    -------
    float
        The upper bound of the truncated Gumbel (max.) distribution.
    """
    _, _, _, upper_bound = _get_parameters(parameters)

    # Get the upper bound of the untruncated Gumbel distribution
    ub_gumbel = gumbel.upper(parameters[:2])
    if upper_bound > ub_gumbel:
        # Use the corresponding Gumbel upper bound instead
        return ub_gumbel

    return upper_bound


def pdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the PDF values of a truncated Gumbel (max.) distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) of a truncated Gumbel (max.) distribution.
    parameters : ARRAY_FLOAT
        Parameters of the truncated Gumbel (max.) distribution.
    lower_bound : float
        Lower bound of the truncated Gumbel (max.) distribution.
    upper_bound : float
        Upper bound of the truncated Gumbel (max.) distribution.

    Returns
    -------
    ARRAY_FLOAT
        PDF values of the truncated Gumbel (max.) distribution
        on the sample values.

    Notes
    -----
    - The values outside the bounds are set to 0.0.
    - ``lower_bound`` and ``upper_bound`` in the function parameters are the
      same as the ones in ``parameters``. For a truncated Gumbel (max.)
      distribution, the bounds are part of the parameterization.
    """
    # Get the parameters
    mu, beta, _, _ = _get_parameters(parameters)

    # Get the non-zero indices (within the bounds)
    idx_non_zero = np.logical_and(xx >= lower_bound, xx <= upper_bound)

    # Compute the normalizing factor
    normalizing_factor = _compute_normalizing_factor(
        mu, beta, lower_bound, upper_bound
    )

    # Compute the PDF
    yy = np.zeros(xx.shape)
    yy[idx_non_zero] = gumbel_r.pdf(xx[idx_non_zero], loc=mu, scale=beta)
    yy[idx_non_zero] = yy[idx_non_zero] / normalizing_factor

    return yy


def cdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the CDF values of the truncated Gumbel (max.) distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) of a truncated Gumbel (max.) distribution.
    parameters : ARRAY_FLOAT
        Parameters of the truncated Gumbel (max.) distribution.
    lower_bound : float
        Lower bound of the truncated Gumbel (max.) distribution.
    upper_bound : float
        Upper bound of the truncated Gumbel (max.) distribution.

    Returns
    -------
    ARRAY_FLOAT
        CDF values of the truncated Gumbel (max.) distribution
        on the sample values.

    Notes
    -----
    - CDF for sample with values smaller (resp. larger) than the lower bound
      (resp. upper bound) are set to 0.0 (resp. 1.0).
    - ``lower_bound`` and ``upper_bound`` in the function parameters are the
      same as the ones in ``parameters``. For a truncated Gumbel (max.)
      distribution, the bounds are part of the parameterization.
    """
    # Get the parameters
    mu, beta, _, _ = _get_parameters(parameters)

    # Get the relevant indices
    idx_lower = xx < lower_bound
    idx_upper = xx > upper_bound
    idx_rest = np.logical_and(
        np.logical_not(idx_lower), np.logical_not(idx_upper)
    )

    # Compute the normalizing factor
    normalizing_factor = _compute_normalizing_factor(
        mu, beta, lower_bound, upper_bound
    )

    # Compute the CDF
    yy = np.empty(xx.shape)
    yy[idx_lower] = 0.0
    yy[idx_upper] = 1.0
    yy[idx_rest] = gumbel_r.cdf(xx[idx_rest], loc=mu, scale=beta)
    yy[idx_rest] = yy[idx_rest] - gumbel_r.cdf(lower_bound, loc=mu, scale=beta)
    yy[idx_rest] = yy[idx_rest] / normalizing_factor

    return yy


def icdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the inverse CDF values of a truncated Gumbel (max.) distribution.

    Parameters
    ----------
    xx : np.ndarray
        Sample values (realizations) in the [0, 1] domain.
    parameters : np.ndarray
        Parameters of a truncated Gumbel (max.) distribution.
    lower_bound : float
        Lower bound of the truncated Gumbel (max.) distribution.
    upper_bound : float
        Upper bound of the truncated Gumbel (max.) distribution.

    Returns
    -------
    ARRAY_FLOAT
        Transformed values in the domain of the truncated Gumbel (max.)
        distribution.

    Notes
    -----
    - ICDF for sample with values of 0.0 and 1.0 are automatically set to the
      lower bound and upper bound, respectively.
    - ``lower_bound`` and ``upper_bound`` in the function parameters may not
      be the same as the ones in ``parameters``. If one of the bounds is
      initially set to inf, then it would be automatically set to the bounds
      of the regular Gumbel.
    TODO: values outside [0, 1] must either be an error or NaN
    """

    # Get the parameters
    mu, beta, _, _ = _get_parameters(parameters)

    # Compute the normalizing factor
    # NOTE: Use the processed bounds
    normalizing_factor = _compute_normalizing_factor(
        mu, beta, lower_bound, upper_bound
    )

    # Make sure values outside [0.0, 1.0] are set to NaN
    xx[xx < 0.0] = np.nan
    xx[xx > 1.0] = np.nan

    # Compute the ICDF
    cdf_values = (
        gumbel_r.cdf(lower_bound, loc=mu, scale=beta) + normalizing_factor * xx
    )
    yy = gumbel.icdf(cdf_values, parameters[:2], lower_bound, upper_bound)

    # Check if values are within the set bounds
    yy = postprocess_icdf(yy, lower_bound, upper_bound)

    return yy



"""
Module with routines involving the exponential probability distribution.

The exponential distribution in UQTestFuns is parametrized by a single
parameter: the rate parametere.

The underlying implementation is based on the implementation from scipy.stats.
In the SciPy convention, the rate parameter corresponds to the reciprocal
of the ``scale`` parameter.
"""

import numpy as np
from scipy.stats import expon

from .utils import verify_param_nums, postprocess_icdf
from ....global_settings import ARRAY_FLOAT

DISTRIBUTION_NAME = "exponential"

NUM_PARAMS = 1


def verify_parameters(parameters: ARRAY_FLOAT) -> None:
    """Verify the parameters of an exponential distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameter of an exponential distribution
        (i.e., the rate parameter).

    Returns
    ------
    None
        The function exits without any return value when nothing is wrong.

    Raises
    ------
    ValueError
        If any of the parameter values are invalid
        or the shapes are inconsistent.
    """
    # Verify overall shape
    verify_param_nums(parameters.size, NUM_PARAMS, DISTRIBUTION_NAME)

    if parameters[0] <= 0.0:
        raise ValueError(
            f"The corresponding rate parameter {parameters[0]}"
            f"must be larger than 0.0!"
        )


def lower(parameters: ARRAY_FLOAT) -> float:
    """Get the lower bound of an exponential distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameter of an exponential distribution.

    Returns
    -------
    float
        The lower bound of the exponential distribution.

    Notes
    -----
    - The parameters are not used in determining the lower bound of
      the distribution; it must, however, appear for interface consistency.
      The lower bound of a lognormal distribution is finite; it is 0.0.
    """
    lower_bound = 0.0

    return lower_bound


def upper(parameters: ARRAY_FLOAT) -> float:
    """Get the upper bound of an exponential distribution.

    Parameters
    ----------
    parameters : ARRAY_FLOAT
        The parameters of an exponential distribution

    Returns
    -------
    float
        The upper bound of the exponential distribution.

    Notes
    -----
    - Strictly speaking, an exponential distribution is unbounded on the right.
      However, for numerical reason an upper bound is set.
    - The upper bound of the exponential distribution is chosen such that
      tbe probability mass between the lower and upper bound is at least
      1 - 1e-15.
    """
    # 36.7368005696771 is the quantile values with probability of 1-1e-16
    # for the exponential distribution with rate parameter value 1.0
    upper_bound = float(36.7368005696771 / parameters[0])

    return upper_bound


def pdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the PDF values of an exponential distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) of an exponential distribution.
    parameters : ARRAY_FLOAT
        Parameters of the exponential distribution.
    lower_bound : float
        Lower bound of the exponential distribution.
    upper_bound : float
        Upper bound of the exponential distribution.

    Returns
    -------
    np.ndarray
        PDF values of the exponential distribution on the sample values.

    Notes
    -----
    - The PDF for sample with values outside the bounds are set to 0.0.
    """
    yy = np.zeros(xx.shape)
    idx = np.logical_and(xx >= lower_bound, xx <= upper_bound)
    rate = parameters[0]
    scale = 1 / rate
    yy[idx] = expon.pdf(xx[idx], scale=scale)

    return yy


def cdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the CDF values of an exponential distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) of an exponential distribution.
    parameters : ARRAY_FLOAT
        Parameters of the exponential distribution.
    lower_bound : float
        Lower bound of the exponential distribution.
    upper_bound : float
        Upper bound of the exponential distribution.

    Returns
    -------
    ARRAY_FLOAT
        CDF values of the exponential distribution on the sample values.

    Notes
    -----
    - The CDF for sample with values smaller (resp. larger)
      than the lower bound (resp. upper bound) are set to 0.0 (resp. 1.0).
    """
    yy = np.empty(xx.shape)
    idx_lower = xx < lower_bound
    idx_upper = xx > upper_bound
    idx_rest = np.logical_and(
        np.logical_not(idx_lower), np.logical_not(idx_upper)
    )

    yy[idx_lower] = 0.0
    yy[idx_upper] = 1.0
    rate = parameters[0]
    scale = 1 / rate
    yy[idx_rest] = expon.cdf(xx[idx_rest], scale=scale)

    return yy


def icdf(
    xx: ARRAY_FLOAT,
    parameters: ARRAY_FLOAT,
    lower_bound: float,
    upper_bound: float,
) -> ARRAY_FLOAT:
    """Get the inverse CDF values of an exponential distribution.

    Parameters
    ----------
    xx : ARRAY_FLOAT
        Sample values (realizations) in the [0, 1] domain.
    parameters : ARRAY_FLOAT
        Parameters of an exponential distribution.
    lower_bound : float
        Lower bound of the exponential distribution.
    upper_bound : float
        Upper bound of the exponential distribution.

    Returns
    -------
    ARRAY_FLOAT
        Transformed values in the domain of the normal distribution.

    Notes
    -----
    - ICDF for sample values outside [0.0, 1.0] is set to NaN according to
      the underlying SciPy implementation.
    """

    # Compute the ICDF
    rate = parameters[0]
    scale = 1 / rate
    yy = expon.ppf(xx, scale=scale)

    # Check if values are within the set bounds
    yy = postprocess_icdf(yy, lower_bound, upper_bound)

    return yy



"""
Module with an implementation of the four-branch test function.

The two-dimensional four-branch function introduced in [1] is a reliability
benchmark problem (see, for instance, [2], [3], [4], [5]).
The test function describes the failure of a series system with four distinct
performance function components.

References
----------

1. Satoshi Katsuki and Dan M. Frangopol, “Hyperspace division method for
   structural Reliability,” Journal of  Engineering Mechanic, vol. 120, no. 11,
   pp. 2405–2427, 1994.
   DOI: 10.1061/(ASCE)0733-9399(1994)120:11(2405)
2. Paul Hendrik Waarts, “Structural reliability using finite element
   analysis - an appraisal of DARS: Directional adaptive response surface
   sampling," Civil Engineering and Geosciences, TU Delft, Delft,
   The Netherlands, 2000.
3. Luc Schueremans and Dionys Van Gemert, “Benefit of splines and neural
   networks in simulation based structural reliability analysis,” Structural
   Safety, vol. 27, no. 3, pp. 246–261, 2005.
   DOI: 10.1016/j.strusafe.2004.11.001
4. B. Echard, N. Gayton, and M. Lemaire, “AK-MCS: An active learning
   reliability method combining Kriging and Monte Carlo Simulation,”
   Structural Safety, vol. 33, no. 2, pp. 145–154, 2011.
   DOI: 10.1016/j.strusafe.2011.01.002
5. Roland Schöbi, Bruno Sudret, and Stefano Marelli, “Rare event estimation
   using polynomial-chaos kriging,” ASCE-ASME Journal Risk and Uncertainty
   in Engineering System, Part A: Civil Engineering, vol. 3, no. 2,
   p. D4016002, 2017.
   DOI: 10.1061/AJRUA6.0000870.
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs, FunParamSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["FourBranch"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Katsuki1994": {
        "function_id": "FourBranch",
        "description": (
            "Input model for the four-branch function "
            "from Katsuki and Frangopol (1994)"
        ),
        "marginals": [
            {
                "name": "X1",
                "distribution": "normal",
                "parameters": [0, 1],
                "description": None,
            },
            {
                "name": "X2",
                "distribution": "normal",
                "parameters": [0, 1],
                "description": None,
            },
        ],
        "copulas": None,
    },
}


AVAILABLE_PARAMETERS: FunParamSpecs = {
    "Katsuki1994": {
        "function_id": "FourBranch",
        "description": (
            "Parameter set for the Four-branch function from Katsuki and "
            "Frangopol (1994)"
        ),
        "declared_parameters": [
            {
                "keyword": "p",
                "value": 3.5 * np.sqrt(2),
                "type": float,
                "description": None,
            },
        ],
    },
    "Schueremans2005": {
        "function_id": "FourBranch",
        "description": (
            "Parameter set for the Four-branch function from Schueremans "
            "(2005)"
        ),
        "declared_parameters": [
            {
                "keyword": "p",
                "value": 6.0 / np.sqrt(2),
                "type": float,
                "description": None,
            },
        ],
    },
}


def evaluate(xx: np.ndarray, p: float) -> np.ndarray:
    """Evaluate the four-branch function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        A two-Dimensional input values given by an N-by-2 array
        where N is the number of input values.
    p : float
        The parameter of the test function; a single float.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """

    # Compute the performance function components
    yy_1 = (
        3.0
        + 0.1 * (xx[:, 0] - xx[:, 1]) ** 2
        - (xx[:, 0] + xx[:, 1]) / np.sqrt(2)
    )
    yy_2 = (
        3.0
        + 0.1 * (xx[:, 0] - xx[:, 1]) ** 2
        + (xx[:, 0] + xx[:, 1]) / np.sqrt(2)
    )
    yy_3 = xx[:, 0] - xx[:, 1] + p
    yy_4 = -1 * xx[:, 0] + xx[:, 1] + p

    yy = np.vstack((yy_1, yy_2, yy_3, yy_4))

    return np.min(yy, axis=0)


class FourBranch(UQTestFunFixDimABC):
    """A concrete implementation of the four-branch test function."""

    _tags = ["reliability"]
    _description = (
        "Series system reliability from Katsuki and Frangopol (1994)"
    )
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = AVAILABLE_PARAMETERS
    _default_parameters_id = "Schueremans2005"

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with an implementation of the speed reducer shaft test function.

The speed reducer shaft test function is a five-dimensional scalar-valued
test function introduced in [1]. It is used as a test function for reliability
analysis algorithms (see, for instance, [1], [2]).

The function models the performance of a shaft in a speed reducer.
The performance is defined as the strength of the shaft subtracted by the
stress. If the value is negative, then the system is in failed state.

References
----------
1. Xiaoping Du and Agus Sudjianto, “First order saddlepoint approximation
   for reliability analysis,” AIAA Journal, vol. 42, no. 6, pp. 1199–1207,
   2004.
   DOI: 10.2514/1.3877
2. X. Li, C. Gong, L. Gu, W. Gao, Z. Jing, and H. Su, “A sequential surrogate
   method for reliability analysis based on radial basis function,”
   Structural Safety, vol. 73, pp. 42–53, 2018.
   DOI: 10.1016/j.strusafe.2018.02.005
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC
from .utils import gumbel_max_mu, gumbel_max_beta

__all__ = ["SpeedReducerShaft"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Du2004": {
        "function_id": "SpeedReducerShaft",
        "description": (
            "Input model for the speed reducer shaft problem "
            "from Du and Sudjianto (2004)"
        ),
        "marginals": [
            {
                "name": "D",
                "distribution": "normal",
                "parameters": [39, 0.1],
                "description": "Shaft diameter [mm]",
            },
            {
                "name": "L",
                "distribution": "normal",
                "parameters": [400, 0.1],
                "description": "Shaft span [mm]",
            },
            {
                "name": "F",
                "distribution": "gumbel",
                "parameters": [gumbel_max_mu(1500, 350), gumbel_max_beta(350)],
                "description": "External force [N]",
            },
            {
                "name": "T",
                "distribution": "normal",
                "parameters": [250, 35],
                "description": "Torque [Nm]",
            },
            {
                "name": "S",
                "distribution": "uniform",
                "parameters": [70, 80],
                "description": "Strength [MPa]",
            },
        ],
        "copulas": None,
    },
}


def evaluate(xx: np.ndarray) -> np.ndarray:
    """Evaluate the speed reducer shaft test function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        A five-dimensional input values given by N-by-5 arrays
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the speed reducer shaft test function.
        If negative, then the system is in failed state.
        The output is a one-dimensional array of length N.
    """
    dd = xx[:, 0]  # diameter [mm]
    ll = xx[:, 1]  # span [mm]
    ff = xx[:, 2]  # external force [N]
    tt = xx[:, 3]  # torque [Nm]
    ss = xx[:, 4]  # strength [MPa]

    # NOTE: Convert [MPa] to [Pa] and [mm] to [m]
    yy = ss * 1e6 - (32 / np.pi / (dd / 1e3) ** 3) * np.sqrt(
        ff**2 * (ll / 1e3) ** 2 / 16 + tt**2
    )

    return yy


class SpeedReducerShaft(UQTestFunFixDimABC):
    """A concrete implementation of the speed reducer shaft function."""

    _tags = ["reliability"]
    _description = (
        "Reliability of a shaft in a speed reducer "
        "from Du and Sudjianto (2004)"
    )
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None

    evaluate = staticmethod(evaluate)  # type: ignore



"""
This module contains the implementation of the meta in UQTestFuns.

The implementation is based on the meta proposed in [1].
To further specify the meta, a set of default parameter values
according to [1] is also implemented.

References
----------

1. William Becker, “Metafunctions for benchmarking in sensitivity analysis,”
   Reliability Engineering & System Safety, vol. 204, p. 107189, 2020,
   doi:
   `10.1016/j.ress.2020.107189 <https://doi.org/10.1016/j.ress.2020.107189>`_.
"""

import math
import numpy as np

from dataclasses import dataclass
from numpy.typing import ArrayLike
from typing import Optional, Union, List

from uqtestfuns.core.parameters import FunParams
from .metaspec import UQMetaFunSpec, UQTestFunSpec
from .basis_functions import BASIS_BY_ID
from ..core import UQTestFun, ProbInput, Marginal


__all__ = ["UQMetaTestFun", "default_coeffs_gen"]


def default_coeffs_gen(sample_size: int) -> np.ndarray:
    r"""Generate coefficient values for each effect term using the default.

    The default coefficient values generator is a Gaussian mixture distribution
    as proposed in [1] with the following form:

    .. math::
       \mathcal{GM}(\phi, \mu_1, \sigma_1, \mu_2, \sigma_2) =
       \phi \mathcal{N}(\mu_1, \sigma_1)
       + (1 - \phi) \mathcal{N}(\mu_2, \sigma_2)

    where :math:`\phi = 0.7`, :math:`\mu_1 = 0`, :math:`\mu_2 = 0`,
    :math:`\sigma_1 = \sqrt{0.5}`, and :math:`\sigma_2 = \sqrt{5}`.

    Parameters
    ----------
    sample_size : int
        Number of sample points to generate.

    Returns
    -------
    np.ndarray
        The coefficient values generated by the generator.
    """
    yy = np.zeros(sample_size)

    phi = 0.7
    sigma_1 = np.sqrt(0.5)
    sigma_2 = np.sqrt(5)

    choices = np.random.choice([0, 1], p=[phi, 1 - phi], size=sample_size)
    yy[choices == 0] = np.random.normal(0, sigma_1, size=np.sum(choices == 0))
    yy[choices == 1] = np.random.normal(0, sigma_2, size=np.sum(choices == 1))

    return yy


def _evaluate_test_function(xx: np.ndarray, spec: UQTestFunSpec) -> np.ndarray:
    """Alternative way to evaluate metafunction realizations."""

    basis_functions = spec.basis_functions
    selected_basis = spec.selected_basis

    # Evaluate the selected basis functions on xx
    basis_vals = np.zeros((xx.shape[0], spec.input_dimension))
    for idx, basis_id in enumerate(selected_basis):
        basis_vals[:, idx] = basis_functions[basis_id](xx[:, idx])

    effects_tuples = spec.effects_tuples
    effects_coeffs = spec.effects_coeffs

    # Evaluate each effect term of the test function
    yy = np.zeros(xx.shape[0])
    for key in effects_tuples:
        n_way = effects_tuples[key]
        coeffs = effects_coeffs[key]

        for dim_indices, coeff in zip(n_way, coeffs):
            yy[:] += coeff * np.prod(basis_vals[:, dim_indices], axis=1)

    return yy


@dataclass
class UQMetaTestFun:
    """Implementation of the metafunction.

    A metafunction is a test-function-generating function.
    A random realization of a test function can be generated
    from a metafunction.

    Parameters
    ----------
    metafun_spec : UQMetaFunSpec
        An instance of meta specification that fully defines a
        meta.
    """

    metafun_spec: UQMetaFunSpec

    def get_sample(
        self, sample_size=1
    ) -> Optional[Union[UQTestFun, List[UQTestFun]]]:
        """Generate realizations of test function.

        Parameters
        ----------
        sample_size : int
            Number of test function realizations.

        Returns
        -------
        Optional[Union[UQTestFun, List[UQTestFun]]]
            Realization(s) of UQTestFunSpec.

        Notes
        -----
        - With sample_size larger than one, a list of UQTestFun instances
          is returned.
        """
        if sample_size < 1:
            return None

        # Generate realizations of test function specification
        testfun_specs = self.metafun_spec.get_sample(sample_size)

        # Create the name
        name = "metafunction_realization"
        # Create the evaluation callable
        evaluate = _evaluate_test_function

        if sample_size == 1:
            assert testfun_specs is not None
            assert not isinstance(testfun_specs, list)
            # Create an instance of inputs
            prob_input = ProbInput(testfun_specs.inputs)
            # Assign the realized spec as a parameter
            parameters = FunParams(
                declared_parameters=[
                    {
                        "keyword": "spec",
                        "value": testfun_specs,
                        "type": UQTestFunSpec,
                        "description": None,
                    },
                ],
            )

            return UQTestFun(
                evaluate=evaluate,
                prob_input=prob_input,
                parameters=parameters,
                function_id=name,
            )

        sample = []
        for i in range(sample_size):
            # Create an instance of inputs
            assert isinstance(testfun_specs, list)
            prob_input = ProbInput(testfun_specs[i].inputs)
            # Assign the realized spec as a parameter
            parameters = FunParams(
                declared_parameters=[
                    {
                        "keyword": "spec",
                        "value": testfun_specs[i],
                        "type": UQTestFunSpec,
                        "description": None,
                    },
                ],
            )

            sample.append(
                UQTestFun(
                    evaluate=evaluate,
                    prob_input=prob_input,
                    parameters=parameters,
                    function_id=name,
                )
            )

        return sample

    @classmethod
    def from_default(
        cls,
        input_dimension: ArrayLike,
        input_id: Optional[int] = None,
    ):
        """Create a metafunction with parameters according to Becker (2019).

        Parameters
        ----------
        input_dimension : Union[int, Sized]
            Number of dimensions of the test functions generated
            by the meta. If a set of values are given, a single value
            will be selected at random.

        Returns
        -------
        UQMetaTestFun
            An instance of metafunction with the default parameters.
        """
        # Select a single input dimension randomly
        if not isinstance(input_dimension, int):
            input_dimension = np.asarray(input_dimension)
            input_dimension = int(np.random.choice(input_dimension))

        effects_dict = {
            1: None,  # Take all
            2: math.floor(0.5 * input_dimension),
            3: math.floor(0.2 * input_dimension),
        }

        if input_id is None:
            input_id = np.random.randint(0, 8)

        input_marginals = [
            Marginal(distribution="uniform", parameters=[0, 1]),
            Marginal(
                distribution="trunc-normal",
                parameters=[0.5, 0.15, 0.0, 1.0],
            ),
            Marginal(distribution="beta", parameters=[8.0, 2.0, 0.0, 1.0]),
            Marginal(distribution="beta", parameters=[2.0, 8.0, 0.0, 1.0]),
            Marginal(distribution="beta", parameters=[2.0, 0.8, 0.0, 1.0]),
            Marginal(distribution="beta", parameters=[0.8, 2.0, 0.0, 1.0]),
            Marginal(distribution="logitnormal", parameters=[0.0, 3.16]),
        ]

        if input_id < 7:
            input_marginals = [input_marginals[input_id]]

        metafun_spec = UQMetaFunSpec(
            input_dimension=input_dimension,
            basis_functions=BASIS_BY_ID,
            effects_dict=effects_dict,
            input_marginals=input_marginals,
            coeffs_generator=default_coeffs_gen,
        )

        return cls(metafun_spec)



"""
Module with an implementation of the 1D Gramacy (2007) test function.

The Gramacy (2007) one-dimensional sine function is a scalar-valued function
that features two regimes: one part is a mixture of sines and cosines and
another part is a linear function. The function was introduced in [1]
as a test function for non-stationary Gaussian process metamodeling
by partitioning the input parameter space.

In its original usage, the response is disturbed by an i.i.d Gaussian noise.

References
----------

1. Robert B. Gramacy, “tgp: An R Package for Bayesian nonstationary,
   semiparametric nonlinear regression and design by Treed Gaussian Process
   models,” Journal of Statistical Software, vol. 19, no. 9, 2007.
   DOI: 10.18637/jss.v019.i09.
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["Gramacy1DSine"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Gramacy2007": {
        "function_id": "Gramacy1DSine",
        "description": (
            "Input model for the one-dimensional function "
            "from Gramacy (2007)"
        ),
        "marginals": [
            {
                "name": "x",
                "distribution": "uniform",
                "parameters": [0.0, 20.0],
                "description": None,
            },
        ],
        "copulas": None,
    },
}


def evaluate_1dsine(xx: np.ndarray) -> np.ndarray:
    """Evaluate the 1D Gramacy (2007) Sine function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        1-Dimensional input values given by an N-by-1 array
        where N is the number of input values.

    parameters : Generator, optional
        A random number generator to generate the noise.

    Returns
    -------
    np.ndarray
        The output of the 1D Gramacy (2007) function evaluated
        on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = np.zeros(len(xx))
    idx_1 = xx[:, 0] <= 9.6
    idx_2 = xx[:, 0] > 9.6
    yy[idx_1] = np.sin(0.2 * np.pi * xx[idx_1, 0]) + 0.2 * np.cos(
        0.8 * np.pi * xx[idx_1, 0]
    )
    yy[idx_2] = -1 + 0.1 * xx[idx_2, 0]

    return yy


class Gramacy1DSine(UQTestFunFixDimABC):
    """A concrete implementation of the 1D Gramacy (2007) Sine function."""

    _tags = ["metamodeling"]
    _description = "One-dimensional sine function from Gramacy (2007)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None

    evaluate = staticmethod(evaluate_1dsine)  # type: ignore



"""
This module implements the solar cell model from Constantine et al. (2015).

The test function is a five-dimensional, scalar-valued function that models
the maximum power of a single-diode solar cell. The function was used in
Constantine et al. (2015) to demonstrate the active subspace method for
input dimension reduction and sensitivity analysis.

References
----------

1. P. G. Constantine, B. Zaharatos, and M. Campanelli, “Discovering an active
   subspace in a single‐diode solar cell model,” Statistical Analysis, vol. 8,
   no. 5–6, pp. 264–273, 2015.
   DOI: 10.1002/sam.11281
"""

import numpy as np
from scipy.optimize import root, minimize
from typing import Tuple

from uqtestfuns.core.custom_typing import (
    MarginalSpecs,
    FunParamSpecs,
    ProbInputSpecs,
)
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["SolarCell"]


MARGINALS_CONSTANTINE2015: MarginalSpecs = [
    {
        "name": "Isc",
        "distribution": "uniform",
        "parameters": [0.05989, 0.23958],
        "description": "Short-circuit current [A]",
    },
    {
        "name": "log_Is",
        "distribution": "uniform",
        "parameters": [np.log(2.2e-11), np.log(2.2e-7)],
        "description": "(log) Diode reverse saturation current [A]",
    },
    {
        "name": "n",
        "distribution": "uniform",
        "parameters": [1.0, 2.0],
        "description": "Ideality factor [-]",
    },
    {
        "name": "Rs",
        "distribution": "uniform",
        "parameters": [0.16625, 0.66500],
        "description": "Series resistance [Ohm]",
    },
    {
        "name": "Rp",
        "distribution": "uniform",
        "parameters": [93.75, 375.00],
        "description": "Parallel (shunt) resistance [Ohm]",
    },
]

AVAILABLE_INPUTS: ProbInputSpecs = {
    "Constantine2015": {
        "function_id": "SolarCell",
        "description": (
            "Probabilistic input model for the single-diode solar cell model "
            "from Constantine et al. (2015)"
        ),
        "marginals": MARGINALS_CONSTANTINE2015,
        "copulas": None,
    },
}


AVAILABLE_PARAMETERS: FunParamSpecs = {
    "Constantine2015": {
        "function_id": "SolarCell",
        "description": (
            "Parameter set for the single-diode solar cell model from "
            "Constantine et al. (2015)"
        ),
        "declared_parameters": [
            {
                "keyword": "n_s",
                "value": 1,
                "type": int,
                "description": "# of cells connected in series [-]",
            },
            {
                "keyword": "v_th",
                "value": 0.02585,
                "type": float,
                "description": "Thermal voltage at 25degC [V]",
            },
        ],
    },
}


def obj_fun_root(
    i: float,
    v: float,
    x: np.ndarray,
    n_s: int,
    v_th: float,
) -> float:
    """Compute the equation to get the root.

    Parameters
    ----------
    i : float
        The current [A]
    v : float
        The voltage [V]
    x : np.ndarray
        The input variables of the solar cell model: I_SC, I_S, n, R_S, R_P.
    n_s : int
        The number of cells connected in series.
    v_th : float
        The thermal voltage value at 25degC [V]

    Returns
    -------
    float
        The root objective function for implicit current; the function returns
        0 if the current corresponds to the given voltage.
    """
    # Get the input variable values
    i_sc = x[0]
    i_s = np.exp(x[1])
    n = x[2]
    r_s = x[3]
    r_p = x[4]

    # Compute the photo current
    exp_term_1 = np.exp(i_sc * r_s / n_s / n / v_th) - 1
    i_l = i_sc + i_s * exp_term_1 + i_sc * r_s / r_p

    # Compute the objective function
    exp_term_2 = np.exp((v + i * r_s) / n_s / n / v_th) - 1
    y = i - i_l + i_s * exp_term_2 + (v + i * r_s) / r_p

    return y


def compute_current(
    v: float,
    x: np.ndarray,
    n_s: int,
    v_th: float,
    **kwargs,
) -> float:
    """Compute the current of the solar cell given a voltage value.

    Parameters
    ----------
    v : float
        The voltage [V]
    x : np.ndarray
        The input variables of the solar cell model: I_SC, I_S, n, R_S, R_P.
    n_s : int
        The number of cells connected in series.
    v_th : float
        The thermal voltage value at 25degC [V]

    Returns
    -------
    float
        The current that corresponds to the given voltage.
    """
    i0 = np.array([0.0])
    # Find the corresponding current as the root
    sol = root(obj_fun_root, i0, args=(v, x, n_s, v_th), **kwargs)
    i = sol.x

    return i


def compute_power_max(
    xx: np.ndarray,
    n_s: int,
    v_th: float,
    **kwargs,
) -> Tuple[np.ndarray, np.ndarray]:
    """Compute the maximum power a solar cell given its design parameters.

    Parameters
    ----------
    xx : np.ndarray
        A five-dimensional input values of the solar cell model given by
        an N-by-5 array where N is the number of input values.
    n_s : int
        The number of cells connected in series.
    v_th : float
        The thermal voltage value at 25degC [V]
    kwargs : dict
        The additional parameters to be passed either to `root()` or
        `minimize()`. A dictionary value of key `root` will be passed to
        `root()`, while a dictionary value of key `minimize` will be passed to
        `minimize()`

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        A pair of arrays containing the maximum power a solar cell and the
        corresponding voltage.
    """
    # Initialize output array
    pp_max = np.empty(len(xx))
    vv_max = np.empty(len(xx))

    # Get additional arguments for root and minimize
    root_kwargs = kwargs.get("root", {})
    minimize_kwargs = kwargs.get("minimize", {})

    for idx in range(len(xx)):
        # Set up the objective function (negated for minimization)
        x = xx[idx]

        def _obj_fun(v, x, n_s, v_th, root_kwargs):
            return -1 * v * compute_current(v, x, n_s, v_th, **root_kwargs)

        # Minimize the objective function
        v0 = np.array([0.0])
        res = minimize(
            _obj_fun,
            v0,
            args=(x, n_s, v_th, root_kwargs),
            **minimize_kwargs,
        )

        pp_max[idx] = -1 * res.fun  # Negated back for the maximum value
        vv_max[idx] = res.x

    return pp_max, vv_max


def evaluate(xx: np.ndarray, n_s: int, v_th: float, **kwargs) -> np.ndarray:
    """Evaluate the solar cell model on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        A five-dimensional input values given by an N-by-5 array
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy, _ = compute_power_max(xx, n_s, v_th, **kwargs)

    return yy


class SolarCell(UQTestFunFixDimABC):
    """Concrete implementation of the single-diode solar cell model."""

    _tags = ["metamodeling", "sensitivity"]
    _description = (
        "Single-diode solar-cell model from Constantine et al. (2015)"
    )
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = AVAILABLE_PARAMETERS

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with an implementation of the convex failure domain problem.

The two-dimensional function is a reliability analysis problem used,
for instance, in [1] and [2].

References
----------
1. A. Borri and E. Speranzini, “Structural reliability analysis using a
   standard deterministic finite element code,” Structural Safety, vol. 19,
   no. 4, pp. 361–382, 1997. DOI: 10.1016/S0167-4730(97)00017-9
2. Paul Hendrik Waarts, “Structural reliability using finite element
   analysis - an appraisal of DARS: Directional adaptive response surface
   sampling," Civil Engineering and Geosciences, TU Delft, Delft,
   The Netherlands, 2000.
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["ConvexFailDomain"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Borri1997": {
        "function_id": "ConvexFailDomain",
        "description": (
            "Input model for the convex failure domain problem "
            "from Borri and Speranzini (1997)"
        ),
        "marginals": [
            {
                "name": "X1",
                "distribution": "normal",
                "parameters": [0, 1],
                "description": None,
            },
            {
                "name": "X2",
                "distribution": "normal",
                "parameters": [0, 1],
                "description": None,
            },
        ],
        "copulas": None,
    },
}


def evaluate(xx: np.ndarray) -> np.ndarray:
    """Evaluate the convex failure domain function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        A two-dimensional input values given by an N-by-2 array
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The performance function of the problem.
        If negative, the system is in failed state.
        The output is a one-dimensional array of length N.
    """

    # Compute the performance function
    yy = (
        0.1 * (xx[:, 0] - xx[:, 1]) ** 2
        - (xx[:, 0] + xx[:, 1]) / np.sqrt(2)
        + 2.5
    )

    return yy


class ConvexFailDomain(UQTestFunFixDimABC):
    """Concrete implementation of the Convex failure domain reliability."""

    _tags = ["reliability"]
    _description = (
        "Convex failure domain problem from Borri and Speranzini (1997)"
    )
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with an implementation of the Bratley et al. (1992) test functions.

The Bratley et al. (1992) paper [1] contains four M-dimensional scalar-valued
functions for testing multi-dimensional numerical integrations using low
discrepancy sequences. The functions was used in [2] and [3] in the context
of global sensitivity analysis.

The four functions are:

- Bratley1992b: A product of trigonometric function;
  integration test function #2
- Bratley1992d: A sum of products integrand; integration test function #4

The Bratley1992d function may also be referred to as the "Bratley function"
or "Bratley et al. (1992) function" in the literature.

References
----------

1. Paul Bratley, Bennet L. Fox, and Harald Niederreiter, "Implementation and
   tests of low-discrepancy sequences," ACM Transactions on Modeling and
   Computer Simulation, vol. 2, no. 3, pp. 195-213, 1992.
   DOI:10.1145/146382.146385
2. S. Kucherenko, M. Rodriguez-Fernandez, C. Pantelides, and N. Shah,
   “Monte Carlo evaluation of derivative-based global sensitivity measures,”
   Reliability Engineering & System Safety, vol. 94, pp. 1137–1148, 2009.
   DOI:10.1016/j.ress.2008.05.006
3. A. Saltelli, P. Annoni, I. Azzini, F. Campolongo, M. Ratto,
   and S. Tarantola, “Variance based sensitivity analysis of model output.
   Design and estimator for the total sensitivity index,” Computer Physics
   Communications, vol. 181, no. 2, pp. 259–270, 2010,
   DOI:10.1016/j.cpc.2009.09.018
"""

import numpy as np

from scipy.special import eval_chebyt

from .sobol_g import evaluate as evaluate_sobol_g

from uqtestfuns.core.custom_typing import ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunVarDimABC

__all__ = ["Bratley1992a", "Bratley1992b", "Bratley1992c", "Bratley1992d"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Bratley1992": {
        "function_id": "Bratley1992",
        "description": (
            "Integration domain of the functions from Bratley et al. (1992)"
        ),
        "marginals": [
            {
                "name": "X",
                "distribution": "uniform",
                "parameters": [0.0, 1.0],
                "description": None,
            },
        ],
        "copulas": None,
    },
}

# Common metadata used in each class definition of Bratley test functions
COMMON_METADATA = dict(
    _tags=["integration", "sensitivity"],
    _available_inputs=AVAILABLE_INPUTS,
    _available_parameters=None,
    _description="from Bratley et al. (1992)",
)


def evaluate_bratley1992a(xx: np.ndarray) -> np.ndarray:
    """Evaluate the test function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        M-Dimensional input values given by an N-by-M array where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    # The function is the Sobol'-G function with all the parameters set to 0.
    parameters = np.zeros(xx.shape[1])
    yy = evaluate_sobol_g(xx, parameters)

    return yy


class Bratley1992a(UQTestFunVarDimABC):
    """An implementation of the test function 1 from Bratley et al. (1992).

    The function (used as an integrand) is a product of an absolute function.

    Notes
    -----
    - The function is the Sobol'-G function with all the parameters set to 0.
    """

    _tags = COMMON_METADATA["_tags"]
    _description = (
        f"Integration test function #1 {COMMON_METADATA['_description']}"
    )
    _available_inputs = COMMON_METADATA["_available_inputs"]
    _available_parameters = COMMON_METADATA["_available_parameters"]

    evaluate = staticmethod(evaluate_bratley1992a)  # type: ignore


def evaluate_bratley1992b(xx: np.ndarray) -> np.ndarray:
    """Evaluate the test function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        M-Dimensional input values given by an N-by-M array where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """

    num_dim = xx.shape[1]

    # Compute the function
    ii = np.arange(1, num_dim + 1)
    yy = np.prod(ii * np.cos(ii * xx), axis=1)

    return yy


class Bratley1992b(UQTestFunVarDimABC):
    """An implementation of the test function 2 from Bratley et al. (1992).

    The function (used as an integrand) is a product of a trigonometric
    function.
    """

    _tags = COMMON_METADATA["_tags"]
    _description = (
        f"Integration test function #2 {COMMON_METADATA['_description']}"
    )
    _available_inputs = COMMON_METADATA["_available_inputs"]
    _available_parameters = COMMON_METADATA["_available_parameters"]

    evaluate = staticmethod(evaluate_bratley1992b)  # type: ignore


def evaluate_bratley1992c(xx: np.ndarray) -> np.ndarray:
    """Evaluate the test function #3 of Bratley et al. (1992).

    Parameters
    ----------
    xx : np.ndarray
        M-Dimensional input values given by an N-by-M array where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """

    num_points, num_dim = xx.shape
    yy = np.ones(num_points)

    # Compute the function
    for m in range(1, num_dim + 1):
        mi = m % 4 + 1
        yy *= eval_chebyt(mi, 2 * xx[:, m - 1] - 1)

    return yy


class Bratley1992c(UQTestFunVarDimABC):
    """An implementation of the test function 3 from Bratley et al. (1992).

    The function (used as an integrand) is a product of a trigonometric
    function.
    """

    _tags = COMMON_METADATA["_tags"]
    _description = (
        f"Integration test function #3 {COMMON_METADATA['_description']}"
    )
    _available_inputs = COMMON_METADATA["_available_inputs"]
    _available_parameters = COMMON_METADATA["_available_parameters"]

    evaluate = staticmethod(evaluate_bratley1992c)  # type: ignore


def evaluate_bratley1992d(xx: np.ndarray) -> np.ndarray:
    """Evaluate the test function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        M-Dimensional input values given by an N-by-M array where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """

    num_points, num_dim = xx.shape
    yy = np.zeros(num_points)

    # Compute the function
    for j in range(num_dim):
        yy += (-1) ** (j + 1) * np.prod(xx[:, : j + 1], axis=1)

    return yy


class Bratley1992d(UQTestFunVarDimABC):
    """An implementation of the test function 4 from Bratley et al. (1992).

    The function (used as an integrand) is a sum of products.
    """

    _tags = COMMON_METADATA["_tags"]
    _description = (
        f"Integration test function #4 {COMMON_METADATA['_description']}"
    )
    _available_inputs = COMMON_METADATA["_available_inputs"]
    _available_parameters = COMMON_METADATA["_available_parameters"]

    evaluate = staticmethod(evaluate_bratley1992d)  # type: ignore



"""
Utility module used across package.
"""

import os
import inspect

from importlib import import_module
from typing import List, Optional, Tuple, Any
from types import ModuleType

SUPPORTED_TAGS = [
    "sensitivity",
    "optimization",
    "metamodeling",
    "reliability",
    "integration",
]


def get_available_classes(
    package: ModuleType, exclude: Optional[List[str]] = None
) -> List[Tuple[str, Any]]:
    """Get the available classes within a given package.

    Parameters
    ----------
    package : ModuleType
        Fully-qualified package name whose contents (all modules)
        are searched through.
    exclude : Optional[List[str]]
        List of modules within package to exclude (don't include '.py').

    Returns
    -------
    List[Tuple[str, str]]
        List of tuples each element of which is the found class name and
        the fully-qualified class path.
    """

    if exclude is None:
        exclude = []

    # Verify package
    err_msg = f"Invalid package name {package}!"
    try:
        package_file = package.__file__
        if package_file is None:
            raise ValueError(err_msg)
    except AttributeError:
        raise AttributeError(err_msg)

    # Get the full path of a (sub-) package
    package_dir = os.path.dirname(package_file)

    # Get all modules within the package
    all_modules = os.listdir(package_dir)

    classes = []
    for module in all_modules:
        if module.endswith(".py"):
            # Create the full module name
            module_name = f"{package.__name__}.{module.replace('.py', '')}"
            if module_name not in exclude:
                all_members = import_module(module_name)
                all_classes = inspect.getmembers(all_members, inspect.isclass)
                for class_name, class_path in all_classes:
                    if class_path.__module__ == module_name:
                        # This is the class that is defined in the module
                        classes.append((class_name, class_path))

    return classes



"""
Module with implementations of test functions from Oakley and O'Hagan (2002).

The 1D test function from Oakley and O'Hagan (2002) (or `Oakley1D` function
for short) is a one-dimensional scalar-valued function.
It was used in [1] as a test function for illustrating metamodeling
and uncertainty propagation approaches.

References
----------

1. Jeremy Oakley and Anthony O'Hagan, "Bayesian inference for the uncertainty
   distribution of computer model outputs," Biometrika , Vol. 89, No. 4,
   p. 769-784, 2002.
   DOI: 10.1093/biomet/89.4.769
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["Oakley1D"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Oakley2002": {
        "function_id": "Oakley1D",
        "description": (
            "Probabilistic input model for the one-dimensional function "
            "from Oakley and O'Hagan (2002)"
        ),
        "marginals": [
            {
                "name": "x",
                "distribution": "normal",
                "parameters": [0.0, 4.0],
                "description": None,
            },
        ],
        "copulas": None,
    },
}


def evaluate(xx: np.ndarray) -> np.ndarray:
    """Evaluate the 1D Oakley-O'Hagan function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        1-Dimensional input values given by an N-by-1 array
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the 1D Oakley-O'Hagan function evaluated
        on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = 5 + xx[:, 0] + np.cos(xx[:, 0])

    return yy


class Oakley1D(UQTestFunFixDimABC):
    """An implementation of the 1D function from Oakley & O'Hagan (2002)."""

    _tags = ["metamodeling"]
    _description = "One-dimensional function from Oakley and O'Hagan (2002)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Utility module for all the UQ test functions.
"""

from .prob_input.marginal import Marginal
from .prob_input.probabilistic_input import ProbInput


def create_canonical_uniform_input(
    input_dimension: int, min_value: float, max_value: float
) -> ProbInput:
    """Create a MultivariateInput in a canonical domain of [-1, 1]^M.

    Parameters
    ----------
    input_dimension : int
        The requested number of dimension.
    min_value : float, optional
        The minimum value of the domain.
    max_value : float, optional
        The maximum value of the domain.

    Returns
    -------
    ProbInput
        The M-dimensional MultivariateInput with independent marginals each
        on [min_value, max_value].
    """

    marginals = []

    for i in range(input_dimension):
        marginals.append(
            Marginal(
                name=f"X{i+1}",
                distribution="uniform",
                parameters=[min_value, max_value],
            )
        )

    return ProbInput(marginals)



"""
Module with an implementation of the flood model.

The flood model from [1] is an eight-dimensional scalar-valued function that
computes the maximum annual underflow of a river (in [m]).
A negative value indicates that an overflow (flooding) occurs.
The model is used in the context of sensitivity analysis in [1] and [2]
and has become a canonical example of the OpenTURNS package [3].

The model is based on a simplification of the one-dimensional hydro-dynamical
equations of St. Venant under the assumption of uniform and constant flow rate
and a large rectangular section.

References
----------
1. B. Iooss and P. Lemaître, “A Review on Global Sensitivity Analysis
   Methods,” in Uncertainty Management in Simulation-Optimization of
   Complex Systems, vol. 59, G. Dellino and C. Meloni, Eds.
   Boston, MA: Springer US, 2015, pp. 101–122.
   DOI: 10.1007/978-1-4899-7547-8_5
2. M. Lamboni, B. Iooss, A.-L. Popelin, and F. Gamboa, “Derivative-based
   global sensitivity measures: General links with Sobol’ indices
   and numerical tests,” Mathematics and Computers in Simulation, vol. 87,
   pp. 45–54, 2013.
   DOI: 10.1016/j.matcom.2013.02.002
3. M. Baudin, A. Dutfoy, B. Iooss, and A.-L. Popelin, “OpenTURNS:
   An Industrial Software for Uncertainty Quantification in Simulation,”
   in Handbook of Uncertainty Quantification, R. Ghanem, D. Higdon,
   and H. Owhadi, Eds. Cham: Springer International Publishing, 2017,
   pp. 2001–2038.
   DOI: 10.1007/978-3-319-12385-1_64
"""

import numpy as np

from uqtestfuns.core.custom_typing import MarginalSpecs, ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["Flood"]


MARGINALS_IOOSS2015: MarginalSpecs = [  # From Ref. [1]
    {
        "name": "Q",
        "distribution": "trunc-gumbel",
        "parameters": [1013.0, 558.0, 500.0, 3000.0],
        "description": "Maximum annual flow rate [m^3/s]",
    },
    {
        "name": "Ks",
        "distribution": "trunc-normal",
        "parameters": [30.0, 8.0, 15.0, np.inf],
        "description": "Strickler coefficient [m^(1/3)/s]",
    },
    {
        "name": "Zv",
        "distribution": "triangular",
        "parameters": [49.0, 51.0, 50.0],
        "description": "River downstream level [m]",
    },
    {
        "name": "Zm",
        "distribution": "triangular",
        "parameters": [54.0, 56.0, 55.0],
        "description": "River upstream level [m]",
    },
    {
        "name": "Hd",
        "distribution": "uniform",
        "parameters": [7.0, 9.0],
        "description": "Dyke height [m]",
    },
    {
        "name": "Cb",
        "distribution": "triangular",
        "parameters": [55.0, 56.0, 55.5],
        "description": "Bank level [m]",
    },
    {
        "name": "L",
        "distribution": "triangular",
        "parameters": [4990.0, 5010.0, 5000.0],
        "description": "Length of the river stretch [m]",
    },
    {
        "name": "B",
        "distribution": "triangular",
        "parameters": [295.0, 305.0, 300.0],
        "description": "River width [m]",
    },
]

AVAILABLE_INPUTS: ProbInputSpecs = {
    "Iooss2015": {
        "function_id": "Flood",
        "description": (
            "Probabilistic input model for the Flood model "
            "from Iooss and Lemaître (2015)"
        ),
        "marginals": MARGINALS_IOOSS2015,
        "copulas": None,
    },
}


def evaluate(xx: np.ndarray) -> np.ndarray:
    """Evaluate the flood model test function on a set of input values.

    Parameters
    ----------
    xx : np.ndarray
        A six-dimensional input values given by an N-by-8 array
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the flood model test function, i.e.,
        the height of a river.
        The output is a one-dimensional array of length N.
    """
    qq = xx[:, 0]  # Maximum annual flow rate
    kk_s = xx[:, 1]  # Strickler coefficient
    zz_v = xx[:, 2]  # River downstream level
    zz_m = xx[:, 3]  # River upstream level
    hh_d = xx[:, 4]  # Dyke height
    cc_b = xx[:, 5]  # Bank level
    ll = xx[:, 6]  # Length of the river stretch
    bb = xx[:, 7]  # River width

    # Compute the maximum annual height of the river [m]
    hh_w = (qq / (bb * kk_s * np.sqrt((zz_m - zz_v) / ll))) ** 0.6

    # Compute the maximum annual underflow [m]
    # NOTE: The sign compared to [1] has been inverted below, a negative
    # value indicates an overflow
    ss = cc_b + hh_d - zz_v - hh_w

    return ss


class Flood(UQTestFunFixDimABC):
    """Concrete implementation of the Flood model test function."""

    _tags = ["metamodeling", "sensitivity"]
    _description = "Flood model from Iooss and Lemaître (2015)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with an implementation of the 2D function from Cheng and Sandu (2010).

The two-dimensional function was used in [1] as an example for metamodeling
using polynomial chaos expansion.

References
----------
1. H. Cheng and A. Sandu, “Collocation least-squares polynomial chaos method,”
   in Proceedings of the 2010 Spring Simulation Multiconference, Orlando,
   Florida: Society for Computer Simulation International, 2010, pp. 1–6.
   DOI: 10.1145/1878537.1878621.
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["Cheng2D"]


AVAILABLE_INPUTS: ProbInputSpecs = {
    "Cheng2010": {
        "function_id": "Cheng2D",
        "description": (
            "Probabilistic input model for the 2D test function "
            "from Cheng and Sandu (2010)"
        ),
        "marginals": [
            {
                "name": "X1",
                "distribution": "uniform",
                "parameters": [0.0, 1.0],
                "description": None,
            },
            {
                "name": "X2",
                "distribution": "uniform",
                "parameters": [0.0, 1.0],
                "description": None,
            },
        ],
        "copulas": None,
    },
}


def evaluate(xx: np.ndarray) -> np.ndarray:
    """Evaluate the 2D test function from Cheng and Sandu (2010).

    Parameters
    ----------
    xx : np.ndarray
        A two-dimensional input values given by an N-by-2 array
        where N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    yy = np.cos(np.sum(xx, axis=1)) * np.exp(np.prod(xx, axis=1))

    return yy


class Cheng2D(UQTestFunFixDimABC):
    """Concrete implementation of the function from Cheng and Sandu (2010)."""

    _tags = ["metamodeling"]
    _description = "Two-dimensional test function from Cheng and Sandu (2010)"
    _available_inputs = AVAILABLE_INPUTS
    _available_parameters = None

    evaluate = staticmethod(evaluate)  # type: ignore



"""
Module with implementations of the two functions from Lim et al. (2002).

The test functions are two-dimensional scalar-valued functions used in [1]
to illustrate the connection between Gaussian process metamodels and
polynomials.

References
----------

1. Y. B. Lim, J. Sacks, W. J. Studden, and W. J. Welch, “Design and analysis of
   computer experiments when the output is highly correlated over the input
   space,” Canadian Journal of Statistics, vol. 30, no. 1, pp. 109–126, 2002,
   DOI: 10.2307/3315868
"""

import numpy as np

from uqtestfuns.core.custom_typing import ProbInputSpecs, MarginalSpecs
from uqtestfuns.core.uqtestfun_abc import UQTestFunFixDimABC

__all__ = ["LimPoly", "LimNonPoly"]


MARGINALS_LIM2002: MarginalSpecs = [
    {
        "name": "x1",
        "distribution": "uniform",
        "parameters": [0.0, 1.0],
        "description": None,
    },
    {
        "name": "x2",
        "distribution": "uniform",
        "parameters": [0.0, 1.0],
        "description": None,
    },
]


def evaluate_poly(xx: np.ndarray) -> np.ndarray:
    """Evaluate the polynomial test function from Lim et al (2002).

    The function is a polynomial with maximum total degree of 5.

    Parameters
    ----------
    xx : np.ndarray
        Two-Dimensional input values given by N-by-2 arrays where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """

    trm_1 = 5.0 / 2.0 * xx[:, 0]
    trm_2 = 35.0 / 2.0 * xx[:, 1]
    trm_3 = 5.0 / 2.0 * xx[:, 0] * xx[:, 1]
    trm_4 = 19 * xx[:, 1] ** 2
    trm_5 = 15.0 / 2.0 * xx[:, 0] ** 3
    trm_6 = 5.0 / 2.0 * xx[:, 0] * xx[:, 1] ** 2
    trm_7 = 11.0 / 2.0 * xx[:, 1] ** 4
    trm_8 = xx[:, 0] ** 3 * xx[:, 1] ** 2

    yy = 9 + trm_1 - trm_2 + trm_3 + trm_4 - trm_5 - trm_6 - trm_7 + trm_8

    return yy


class LimPoly(UQTestFunFixDimABC):
    """An implementation of the 2D polynomial from Lim et al. (2002)."""

    _tags = ["metamodeling"]
    _description = "Two-dimensional polynomial function from Lim et al. (2002)"
    _available_inputs: ProbInputSpecs = {
        "Lim2002": {
            "function_id": "LimPoly",
            "description": (
                "Input specification for the polynomial function "
                "from Lim et al. (2002)"
            ),
            "marginals": MARGINALS_LIM2002,
            "copulas": None,
        }
    }
    _available_parameters = None

    evaluate = staticmethod(evaluate_poly)  # type: ignore


def evaluate_non_poly(xx: np.ndarray) -> np.ndarray:
    """Evaluate the non-polynomial test function from Lim et al (2002).

    The function is a polynomial with maximum total degree of 5.

    Parameters
    ----------
    xx : np.ndarray
        Two-Dimensional input values given by N-by-2 arrays where
        N is the number of input values.

    Returns
    -------
    np.ndarray
        The output of the test function evaluated on the input values.
        The output is a 1-dimensional array of length N.
    """
    trm_1 = 30 + 5 * xx[:, 0] * np.sin(5 * xx[:, 0])
    trm_2 = 4 + np.exp(-5 * xx[:, 1])

    yy = (trm_1 * trm_2 - 100) / 6.0

    return yy


class LimNonPoly(UQTestFunFixDimABC):
    """An implementation of the 2D non-polynomial from Lim et al. (2002)."""

    _tags = ["metamodeling"]
    _description = (
        "Two-dimensional non-polynomial function from Lim et al. (2002)"
    )
    _available_inputs: ProbInputSpecs = {
        "Lim2002": {
            "function_id": "LimNonPoly",
            "description": (
                "Input specification for the non-polynomial function "
                "from Lim et al. (2002)"
            ),
            "marginals": MARGINALS_LIM2002,
            "copulas": None,
        }
    }
    _available_parameters = None

    evaluate = staticmethod(evaluate_non_poly)  # type: ignore
